@@ -18,7 +18,7 @@
 
 #include <stdio.h>
 #include <sys/types.h>
-#include <sys/stat.h>
+#include "../posixstat.h"
 #if defined (HAVE_VFPRINTF)
 #include <varargs.h>
 #endif /* VFPRINTF */
@@ -35,11 +35,22 @@
 #include "../jobs.h"
 #include "../builtins.h"
 #include "../input.h"
+#include "../execute_cmd.h"
 #include "hashcom.h"
+#include "common.h"
+#include <tilde/tilde.h>
 
-void no_args (), remember_args (), parse_and_execute_cleanup ();
-void set_dollar_vars_changed (), set_dollar_vars_unchanged ();
-extern int follow_symbolic_links, interactive;
+#if defined (HISTORY)
+#  include "../bashhist.h"
+#endif
+
+extern int no_symbolic_links, interactive, interactive_shell;
+extern int indirection_level, startup_state;
+extern int hashing_disabled;
+extern int variable_context;
+extern char *this_command_name, *shell_name;
+extern COMMAND *global_command;
+extern HASH_TABLE *hashed_filenames;
 
 /* Read a numeric arg for this_command_name, the name of the shell builtin
    that wants it.  LIST is the word list that the arg is to come from. */
@@ -101,10 +112,10 @@ get_numeric_arg (list)
    instead of shell control structures, and it won't ever exit the
    shell. */
 #if defined (HAVE_VFPRINTF)
+void
 builtin_error (va_alist)
      va_dcl
 {
-  extern char *this_command_name;
   char *format;
   va_list args;
 
@@ -118,11 +129,10 @@ builtin_error (va_alist)
   fprintf (stderr, "\n");
 }
 #else /* !HAVE_VFPRINTF */
+void
 builtin_error (format, arg1, arg2, arg3, arg4, arg5)
      char *format, *arg1, *arg2, *arg3, *arg4, *arg5;
 {
-  extern char *this_command_name;
-
   if (this_command_name && *this_command_name)
     fprintf (stderr, "%s: ", this_command_name);
 
@@ -216,8 +226,6 @@ char *
 find_hashed_filename (filename)
      char *filename;
 {
-  extern HASH_TABLE *hashed_filenames;
-  extern int hashing_disabled;
   register BUCKET_CONTENTS *item;
 
   if (hashing_disabled)
@@ -283,8 +291,6 @@ void
 remove_hashed_filename (filename)
      char *filename;
 {
-  extern HASH_TABLE *hashed_filenames;
-  extern int hashing_disabled;
   register BUCKET_CONTENTS *item;
 
   if (hashing_disabled)
@@ -310,14 +316,10 @@ remove_hashed_filename (filename)
 /*								    */
 /* **************************************************************** */
 
-extern int variable_context;
-
 static WORD_LIST **dollar_arg_stack = (WORD_LIST **)NULL;
 static int dollar_arg_stack_slots = 0;
 static int dollar_arg_stack_index = 0;
 
-void push_dollar_vars (), pop_dollar_vars ();
-
 void
 push_context ()
 {
@@ -359,6 +361,7 @@ pop_dollar_vars ()
   dollar_arg_stack[dollar_arg_stack_index] = (WORD_LIST *)NULL;
 }
 
+void
 dispose_saved_dollar_vars ()
 {
   if (!dollar_arg_stack || !dollar_arg_stack_index)
@@ -391,6 +394,7 @@ set_dollar_vars_changed ()
 
 /* Function called when one of the builtin commands detects a bad
    option. */
+void
 bad_option (s)
      char *s;
 {
@@ -405,7 +409,7 @@ char *
 get_working_directory (for_whom)
      char *for_whom;
 {
-  if (!follow_symbolic_links)
+  if (no_symbolic_links)
     {
       if (the_current_working_directory)
 	free (the_current_working_directory);
@@ -415,20 +419,16 @@ get_working_directory (for_whom)
 
   if (!the_current_working_directory)
     {
-      char *directory, *getwd ();
+      char *directory;
 
-      the_current_working_directory = (char *)xmalloc (MAXPATHLEN);
+      the_current_working_directory = xmalloc (MAXPATHLEN);
       directory = getwd (the_current_working_directory);
       if (!directory)
 	{
-	  extern char *shell_name;
-
 	  if (for_whom && *for_whom)
 	    fprintf (stderr, "%s: ", for_whom);
-	  else if (shell_name && *shell_name)
-	    fprintf (stderr, "%s: ", shell_name);
 	  else
-	    fprintf (stderr, "bash: ");
+	    fprintf (stderr, "%s: ", get_name_for_error ());
 
 	  fprintf (stderr, "could not get current directory: %s\n\r",
 		   the_current_working_directory);
@@ -523,9 +523,6 @@ get_job_spec (list)
 }
 #endif /* JOB_CONTROL */
 
-/* The command name of the currently running function. */
-extern char *this_command_name;
-
 int parse_and_execute_level = 0;
 
 /* How to force parse_and_execute () to clean up after itself. */
@@ -542,46 +539,33 @@ parse_and_execute (string, from_file)
      char *string;
      char *from_file;
 {
-#if defined (HISTORY)
-  extern int remember_on_history;
-#  if defined (BANG_HISTORY)
-  extern int history_expansion_inhibited;
-#  endif /* BANG_HISTORY */
-#endif /* HISTORY */
-  extern int indirection_level;
-#if defined (ONESHOT)
-  extern int startup_state;
-#endif /* ONESHOT */
-  extern COMMAND *global_command;
-  extern char *indirection_level_string ();
-  extern int pop_stream ();
-
   int last_result = EXECUTION_SUCCESS;
-  int code, jump_to_top_level = 0;
+  int code = 0, jump_to_top_level = 0;
   char *orig_string = string;
 
   /* Unwind protect this invocation of parse_and_execute (). */
   begin_unwind_frame ("parse_and_execute_top");
-    unwind_protect_int (parse_and_execute_level);
-    unwind_protect_jmp_buf (top_level);
-    unwind_protect_int (indirection_level);
-    unwind_protect_int (interactive);
+  unwind_protect_int (parse_and_execute_level);
+  unwind_protect_jmp_buf (top_level);
+  unwind_protect_int (indirection_level);
 
 #if defined (HISTORY)
-    unwind_protect_int (remember_on_history);
+  if (interactive_shell)
+    {
+      unwind_protect_int (remember_on_history);
 #  if defined (BANG_HISTORY)
-    unwind_protect_int (history_expansion_inhibited);
+      unwind_protect_int (history_expansion_inhibited);
 #  endif /* BANG_HISTORY */
+    }
 #endif /* HISTORY */
 
-    add_unwind_protect (pop_stream, (char *)NULL);
-    if (orig_string)
-      add_unwind_protect (vfree, orig_string);
+  add_unwind_protect (pop_stream, (char *)NULL);
+  if (orig_string)
+    add_unwind_protect (xfree, orig_string);
   end_unwind_frame ();
 
   parse_and_execute_level++;
   push_stream ();
-  interactive = 0;
   indirection_level++;
 
 #if defined (HISTORY)
@@ -636,8 +620,6 @@ parse_and_execute (string, from_file)
 	  {
 	    if ((command = global_command) != (COMMAND *)NULL)
 	      {
-		extern struct fd_bitmap *new_fd_bitmap ();
-		extern void dispose_fd_bitmap ();
 		struct fd_bitmap *bitmap;
 
 		bitmap = new_fd_bitmap (FD_BITMAP_SIZE);
@@ -647,8 +629,9 @@ parse_and_execute (string, from_file)
 		global_command = (COMMAND *)NULL;
 
 #if defined (ONESHOT)
-	        if (startup_state == 2 && command->type == cm_simple &&
-		    !command->redirects && !command->value.Simple->redirects)
+	        if (startup_state == 2 && *bash_input.location.string == '\0' &&
+		    command->type == cm_simple && !command->redirects &&
+		    !command->value.Simple->redirects)
 		  {
 		    command->flags |= CMD_NO_FORK;
 		    command->value.Simple->flags |= CMD_NO_FORK;
@@ -678,7 +661,13 @@ parse_and_execute (string, from_file)
   run_unwind_frame ("parse_and_execute_top");
 
   if (interrupt_state && parse_and_execute_level == 0)
-    throw_to_top_level ();
+    {
+      /* An interrupt during non-interactive execution in an
+         interactive shell (e.g. via $PROMPT_COMMAND) should
+         not cause the shell to exit. */
+      interactive = interactive_shell;
+      throw_to_top_level ();
+    }
 
   if (jump_to_top_level)
     longjmp (top_level, code);
@@ -704,7 +693,7 @@ builtin_address_internal (name, disabled_okay)
 
       j = shell_builtins[mid].name[0] - name[0];
 
-      if (!j)
+      if (j == 0)
 	j = strcmp (shell_builtins[mid].name, name);
 
       if (j == 0)
@@ -758,6 +747,7 @@ shell_builtin_compare (sbp1, sbp2)
 
 /* Sort the table of shell builtins so that the binary search will work
    in find_shell_builtin. */
+void
 initialize_shell_builtins ()
 {
   qsort (shell_builtins, num_shell_builtins, sizeof (struct builtin),
