@@ -19,16 +19,24 @@
    with Bash; see the file COPYING.  If not, write to the Free Software
    Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
+#include "config.h"	/* includes unistd.h for us */
 #include <stdio.h>
 #include <errno.h>
-#if !defined (errno)
-extern int errno;
-#endif /* !errno */
-#include <sys/types.h>
+#include "bashtypes.h"
 #include <sys/param.h>
+#if defined (_POSIX_VERSION)
+#  if defined (amiga) && defined (USGr4)
+#    define _POSIX_SOURCE
+#  endif
+#  include <signal.h>
+#  if defined (amiga) && defined (USGr4)
+#    undef _POSIX_SOURCE
+#  endif
+#endif /* _POSIX_VERSION */
 #include "filecntl.h"
 #include "bashansi.h"
 #include "shell.h"
+#include <tilde/tilde.h>
 
 #if !defined (USG) || defined (HAVE_RESOURCE)
 #  include <sys/time.h>
@@ -37,10 +45,9 @@ extern int errno;
 #include <sys/times.h>
 #include "maxpath.h"
 
-extern char *tilde_expand ();
-
-/* Forward declarations. */
-extern char *canonicalize_pathname ();
+#if !defined (errno)
+extern int errno;
+#endif /* !errno */
 
 /* Make the functions strchr and strrchr if they do not exist. */
 #if !defined (HAVE_STRCHR)
@@ -98,13 +105,13 @@ xmalloc (size)
 
 char *
 xrealloc (pointer, size)
-     register char *pointer;
+     GENPTR pointer;
      int size;
 {
   char *temp;
 
   if (!pointer)
-    temp = (char *)xmalloc (size);
+    temp = xmalloc (size);
   else
     temp = (char *)realloc (pointer, size);
 
@@ -114,16 +121,14 @@ xrealloc (pointer, size)
   return (temp);
 }
 
-#if !defined (NeXT)
 /* Use this as the function to call when adding unwind protects so we
    don't need to know what free() returns. */
 void
-vfree (string)
+xfree (string)
      char *string;
 {
   free (string);
 }
-#endif /* !NeXT */
 
 /* **************************************************************** */
 /*								    */
@@ -185,6 +190,26 @@ all_digits (string)
   return (1);
 }
 
+/* atol(3) is not universal */
+long
+string_to_long (s)
+     char *s;
+{
+  long ret = 0L;
+  int neg = 0;
+
+  while (s && *s && whitespace (*s))
+    s++;
+  if (*s == '-' || *s == '+')
+    {
+      neg = *s == '-';
+      s++;
+    }
+  for ( ; s && *s && digit (*s); s++)
+    ret = (ret * 10) + digit_value (*s);
+  return (neg ? -ret : ret);
+}
+
 /* Return 1 if this token is a legal shell `identifier'; that is, it consists
    solely of letters, digits, and underscores, and does not begin with a
    digit. */
@@ -194,10 +219,7 @@ legal_identifier (name)
 {
   register char *s;
 
-  if (!name || !*name)
-    return (0);
-
-  if (digit (*name))
+  if (!name || !*name || digit (*name))
     return (0);
 
   for (s = name; s && *s; s++)
@@ -208,6 +230,29 @@ legal_identifier (name)
   return (1);
 }
 
+/* Make sure that WORD is a valid shell identifier, i.e.
+   does not contain a dollar sign, nor is quoted in any way.  Nor
+   does it consist of all digits.  If CHECK_WORD is non-zero,
+   the word is checked to ensure that it consists of only letters,
+   digits, and underscores. */
+check_identifier (word, check_word)
+     WORD_DESC *word;
+     int check_word;
+{
+  if (word->dollar_present || word->quoted || all_digits (word->word))
+    {
+      report_error ("`%s' is not a valid identifier", word->word);
+      return (0);
+    }
+  else if (check_word && legal_identifier (word->word) == 0)
+    {
+      report_error ("`%s' is not a valid identifier", word->word);
+      return (0);
+    }
+  else
+    return (1);
+}
+
 /* A function to unset no-delay mode on a file descriptor.  Used in shell.c
    to unset it on the fd passed as stdin.  Should be called on stdin if
    readline gets an EAGAIN or EWOULDBLOCK when trying to read input. */
@@ -284,7 +329,7 @@ map_over_words (words, function)
    function to something, or you will lose the chain. */
 GENERIC_LIST *
 reverse_list (list)
-     register GENERIC_LIST *list;
+     GENERIC_LIST *list;
 {
   register GENERIC_LIST *next, *prev = (GENERIC_LIST *)NULL;
 
@@ -301,7 +346,7 @@ reverse_list (list)
 /* Return the number of elements in LIST, a generic list. */
 int
 list_length (list)
-     register GENERIC_LIST *list;
+     GENERIC_LIST *list;
 {
   register int i;
 
@@ -325,6 +370,7 @@ GENERIC_LIST *
 delete_element (list, comparer, arg)
      GENERIC_LIST **list;
      Function *comparer;
+     char *arg;
 {
   register GENERIC_LIST *prev = (GENERIC_LIST *)NULL;
   register GENERIC_LIST *temp = *list;
@@ -342,14 +388,14 @@ delete_element (list, comparer, arg)
 }
 
 /* Find NAME in ARRAY.  Return the index of NAME, or -1 if not present.
-   ARRAY shoudl be NULL terminated. */
+   ARRAY should be NULL terminated. */
 int
 find_name_in_list (name, array)
-     char *name, *array[];
+     char *name, **array;
 {
   int i;
 
-  for (i=0; array[i]; i++)
+  for (i = 0; array[i]; i++)
     if (strcmp (name, array[i]) == 0)
       return (i);
 
@@ -359,17 +405,17 @@ find_name_in_list (name, array)
 /* Return the length of ARRAY, a NULL terminated array of char *. */
 int
 array_len (array)
-     register char **array;
+     char **array;
 {
   register int i;
-  for (i=0; array[i]; i++);
+  for (i = 0; array[i]; i++);
   return (i);
 }
 
 /* Free the contents of ARRAY, a NULL terminated array of char *. */
 void
 free_array (array)
-     register char **array;
+     char **array;
 {
   register int i = 0;
 
@@ -383,7 +429,7 @@ free_array (array)
 /* Allocate and return a new copy of ARRAY and its contents. */
 char **
 copy_array (array)
-     register char **array;
+     char **array;
 {
   register int i;
   int len;
@@ -409,7 +455,8 @@ list_append (head, tail)
   if (!t_head)
     return (tail);
 
-  while (t_head->next) t_head = t_head->next;
+  while (t_head->next)
+    t_head = t_head->next;
   t_head->next = tail;
   return (head);
 }
@@ -429,7 +476,7 @@ strip_leading (string)
   if (string != start)
     {
       int len = strlen (string);
-      bcopy (string, start, len);
+      FASTCOPY (string, start, len);
       start[len] = '\0';
     }
 }
@@ -557,6 +604,7 @@ make_absolute (string, dot_path)
      char *string, *dot_path;
 {
   char *result;
+  int result_len;
   
   if (!dot_path || *string == '/')
     result = savestring (string);
@@ -564,18 +612,23 @@ make_absolute (string, dot_path)
     {
       if (dot_path && dot_path[0])
 	{
-	  result = (char *)xmalloc (2 + strlen (dot_path) + strlen (string));
+	  result = xmalloc (2 + strlen (dot_path) + strlen (string));
 	  strcpy (result, dot_path);
-	  if (result[strlen (result) - 1] != '/')
-	    strcat (result, "/");
+	  result_len = strlen (result);
+	  if (result[result_len - 1] != '/')
+	    {
+	      result[result_len++] = '/';
+	      result[result_len] = '\0';
+	    }
 	}
       else
 	{
-	  result = (char *)xmalloc (3 + strlen (string));
-	  strcpy (result, "./");
+	  result = xmalloc (3 + strlen (string));
+	  result[0] = '.'; result[1] = '/'; result[2] = '\0';
+	  result_len = 2;
 	}
 
-      strcat (result, string);
+      strcpy (result + result_len, string);
     }
 
   return (result);
@@ -646,27 +699,28 @@ full_pathname (file)
   else
     file = savestring (file);
 
-  if (absolute_pathname (file))
-    if (*file == '/')
-      return (file);
+  if ((*file == '/') && absolute_pathname (file))
+    return (file);
 
   disposer = file;
 
   {
-    char *current_dir = (char *)xmalloc (2 + MAXPATHLEN + strlen (file));
+    char *current_dir = xmalloc (2 + MAXPATHLEN + strlen (file));
+    int dlen;
     if (!getwd (current_dir))
       {
 	report_error (current_dir);
 	free (current_dir);
 	return ((char *)NULL);
       }
-    strcat (current_dir, "/");
+    dlen = strlen (current_dir);
+    current_dir[dlen++] = '/';
 
     /* Turn /foo/./bar into /foo/bar. */
-    if (strncmp (file, "./", 2) == 0)
+    if (file[0] == '.' && file[1] == '/')
       file += 2;
 
-    strcat (current_dir, file);
+    strcpy (current_dir + dlen, file);
     free (disposer);
     return (current_dir);
   }
@@ -676,13 +730,13 @@ full_pathname (file)
    match in s1.  The compare is case insensitive. */
 char *
 strindex (s1, s2)
-     register char *s1, *s2;
+     char *s1, *s2;
 {
   register int i, l = strlen (s2);
   register int len = strlen (s1);
 
   for (i = 0; (len - i) >= l; i++)
-    if (strnicmp (&s1[i], s2, l) == 0)
+    if (strnicmp (s1 + i, s2, l) == 0)
       return (s1 + i);
   return ((char *)NULL);
 }
@@ -750,10 +804,19 @@ set_lines_and_columns (lines, cols)
   free (val);
 }
 
+/* A wrapper for bcopy that can be prototyped in general.h */
+void
+xbcopy (s, d, n)
+     char *s, *d;
+     int n;
+{
+  FASTCOPY (s, d, n);
+}
+
 /* Return a string corresponding to the error number E.  From
    the ANSI C spec. */
 #if defined (strerror)
-#undef strerror
+#  undef strerror
 #endif
 
 #if !defined (HAVE_STRERROR)
@@ -775,7 +838,11 @@ strerror (e)
 }
 #endif /* HAVE_STRERROR */
 
-#if !defined (USG) || defined (HAVE_RESOURCE)
+#if (defined (USG) && !defined (HAVE_TIMEVAL)) || defined (Minix)
+#  define TIMEVAL_MISSING
+#endif
+
+#if !defined (TIMEVAL_MISSING) || defined (HAVE_RESOURCE)
 /* Print the contents of a struct timeval * in a standard way. */
 void
 print_timeval (tvp)
@@ -792,9 +859,9 @@ print_timeval (tvp)
   minutes = seconds / 60;
   seconds %= 60;
 
-  printf ("%0dm%0d.%02ds",  minutes, seconds, seconds_fraction);
+  printf ("%0dm%0ld.%02ds",  minutes, seconds, seconds_fraction);
 }
-#endif
+#endif /* !TIMEVAL_MISSING || HAVE_RESOURCE */
 
 /* Print the time defined by a time_t (returned by the `times' and `time'
    system calls) in a standard way.  This is scaled in terms of HZ, which
@@ -824,7 +891,7 @@ print_time_in_hz (t)
   minutes = seconds / 60;
   seconds %= 60;
 
-  printf ("%0dm%0d.%02ds",  minutes, seconds, seconds_fraction);
+  printf ("%0dm%0ld.%02ds",  minutes, seconds, seconds_fraction);
 }
 #endif /* BrainDeath */
 
@@ -893,7 +960,7 @@ dup2 (fd1, fd2)
 #  define HAVE_GETDTABLESIZE
 #endif /* !USG && !HPUX */
 
-#if defined (hppa) && defined (hpux_8)
+#if defined (hppa) && (defined (hpux_8) || defined (hpux_9))
 #  undef HAVE_GETDTABLESIZE
 #endif /* hppa && hpux_8 */
 
@@ -917,14 +984,14 @@ getdtablesize ()
 }
 #endif /* !HAVE_GETDTABLESIZE */
 
-#if defined (USG) && !defined (sgi)
+#if defined (USG)
 
 #if !defined (HAVE_BCOPY)
-bcopy (s,d,n) char *d,*s; { memcpy (d, s, n); }
+bcopy (s,d,n) char *d,*s; { FASTCOPY (s, d, n); }
 bzero (s,n) char *s; int n; { memset(s, '\0', n); }
 #endif /* !HAVE_BCOPY */
 
-#if !defined (HPUX) && !defined (Linux)
+#if !defined (HAVE_GETHOSTNAME)
 #include <sys/utsname.h>
 int
 gethostname (name, namelen)
@@ -942,8 +1009,8 @@ gethostname (name, namelen)
   name[namelen] = '\0';
   return (0);
 }
-#endif /* !HPUX */
-#endif /* USG && !sgi */
+#endif /* !HAVE_GETHOSTNAME */
+#endif /* USG */
 
 #if !defined (HAVE_GETWD)
 char *
@@ -991,22 +1058,26 @@ int
 getc_with_restart (stream)
      FILE *stream;
 {
-  int result;
-  char c;
+  static char localbuf[128];
+  static int local_index = 0, local_bufused = 0;
 
-  while (1)
+  /* Try local buffering to reduce the number of read(2) calls. */
+  if (local_index == local_bufused || local_bufused == 0)
     {
-      result = read (fileno (stream), &c, sizeof (char));
-
-      if (result == 0)
-	return (EOF);
-
-      if (result == sizeof (char))
-	return (c);
-
-      if (errno != EINTR)
-	return (EOF);
+      while (1)
+	{
+	  local_bufused = read (fileno (stream), localbuf, sizeof(localbuf));
+	  if (local_bufused > 0)
+	    break;
+	  else if (local_bufused == 0 || errno != EINTR)
+	    {
+	      local_index = 0;
+	      return EOF;
+	    }
+	}
+      local_index = 0;
     }
+  return (localbuf[local_index++]);
 }
 #endif /* NO_READ_RESTART_ON_SIGNAL */
 
@@ -1023,12 +1094,9 @@ killpg (pgrp, sig)
      pid_t pgrp;
      int sig;
 {
-  int result;
-
-  result = kill (-pgrp, sig);
-  return (result);
+  return (kill (-pgrp, sig));
 }
-#endif /* USG  || _POSIX_VERSION */
+#endif /* USG  || AIX || (_POSIX_VERSION && Ultrix) */
 
 /* **************************************************************** */
 /*								    */
@@ -1045,10 +1113,13 @@ bash_tilde_expand (text)
 {
   char *result = (char *)NULL;
 
-  if (strcmp (text, "-") == 0)
-    result = get_string_value ("OLDPWD");
-  else if (strcmp (text, "+") == 0)
-    result = get_string_value ("PWD");
+  if (!text[1])
+    {
+      if (*text == '+')
+        result = get_string_value ("PWD");
+      else if (*text == '-')
+        result = get_string_value ("OLDPWD");
+    }
 
   if (result)
     result = savestring (result);
@@ -1062,12 +1133,10 @@ bash_tilde_expand (text)
 void
 tilde_initialize ()
 {
-  extern Function *tilde_expansion_failure_hook;
-  extern char **tilde_additional_prefixes, **tilde_additional_suffixes;
   static int times_called = 0;
 
   /* Tell the tilde expander that we want a crack if it fails. */
-  tilde_expansion_failure_hook = (Function *)bash_tilde_expand;
+  tilde_expansion_failure_hook = (CPFunction *)bash_tilde_expand;
 
   /* Tell the tilde expander about special strings which start a tilde
      expansion, and the special strings that end one.  Only do this once.
@@ -1088,7 +1157,14 @@ tilde_initialize ()
 }
 
 #if defined (_POSIX_VERSION)
-#include <signal.h>
+
+#if !defined (SA_INTERRUPT)
+#  define SA_INTERRUPT 0
+#endif
+
+#if !defined (SA_RESTART)
+#  define SA_RESTART 0
+#endif
 
 SigHandler *
 set_signal_handler (sig, handler)
@@ -1099,6 +1175,12 @@ set_signal_handler (sig, handler)
 
   act.sa_handler = handler;
   act.sa_flags = 0;
+#if 0
+  if (sig == SIGALRM)
+    act.sa_flags |= SA_INTERRUPT;	/* XXX */
+  else
+    act.sa_flags |= SA_RESTART;		/* XXX */
+#endif
   sigemptyset (&act.sa_mask);
   sigemptyset (&oact.sa_mask);
   sigaction (sig, &act, &oact);
