@@ -19,23 +19,19 @@ with Bash; see the file COPYING.  If not, write to the Free Software
 Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include <stdio.h>
-#include <sys/types.h>
+#include "bashtypes.h"
 #include "posixstat.h"
 #include <sys/param.h>
 #include "bashansi.h"
 #include "shell.h"
-
 #include "maxpath.h"
+#include "execute_cmd.h"
+#include <tilde/tilde.h>
 
 #ifndef NOW
 #define NOW ((time_t)time ((time_t *)0))
 #endif
 
-extern struct user_info current_user;
-
-extern char *extract_colon_unit ();
-extern char *tilde_expand ();
-
 typedef struct {
   char *name;
   time_t access_time;
@@ -53,11 +49,12 @@ int mailfiles_count = 0;
 int last_time_mail_checked = 0;
 
 /* Returns non-zero if it is time to check mail. */
+int
 time_to_check_mail ()
 {
   char *temp = get_string_value ("MAILCHECK");
   time_t now = NOW;
-  long seconds = -1;
+  long seconds = -1L;
 
   /* Skip leading whitespace in MAILCHECK. */
   if (temp)
@@ -68,21 +65,19 @@ time_to_check_mail ()
       seconds = atoi (temp);
     }
 
-  /* Negative number, or non-numbers (such as empty string) cause no checking
-     to take place. */
+  /* Negative number, or non-numbers (such as empty string) cause no
+     checking to take place. */
   if (seconds < 0)
     return (0);
 
   /* Time to check if MAILCHECK is explicitly set to zero, or if enough
      time has passed since the last check. */
-  if (!seconds || ((now - last_time_mail_checked) >= seconds))
-    return (1);
-
-  return (0);
+  return (!seconds || ((now - last_time_mail_checked) >= seconds));
 }
 
 /* Okay, we have checked the mail.  Perhaps I should make this function
    go away. */
+void
 reset_mail_timer ()
 {
   last_time_mail_checked = NOW;
@@ -90,59 +85,65 @@ reset_mail_timer ()
 
 /* Locate a file in the list.  Return index of
    entry, or -1 if not found. */
+static int
 find_mail_file (file)
      char *file;
 {
-  register int index = 0;
+  register int i;
 
-  while (index < mailfiles_count)
-    if (strcmp ((mailfiles[index])->name, file) == 0) return index;
-    else index++;
-  return (-1);
+  for (i = 0; i < mailfiles_count; i++)
+    if (STREQ ((mailfiles[i])->name, file))
+      return i;
+
+  return -1;
 }
 
-/* Add this file to the list of remembered files. */
-void
+/* Add this file to the list of remembered files and return its index
+   in the list of mail files. */
+static int
 add_mail_file (file)
      char *file;
 {
   struct stat finfo;
-  char *filename = full_pathname (file);
-  int index = find_mail_file (file);
+  char *filename;
+  int i;
 
-  if (index > -1)
+  filename = full_pathname (file);
+  i = find_mail_file (file);
+  if (i > -1)
     {
       if (stat (filename, &finfo) == 0)
 	{
-	  mailfiles[index]->mod_time = finfo.st_mtime;
-	  mailfiles[index]->access_time = finfo.st_atime;
-	  mailfiles[index]->file_size = (long)finfo.st_size;
+	  mailfiles[i]->mod_time = finfo.st_mtime;
+	  mailfiles[i]->access_time = finfo.st_atime;
+	  mailfiles[i]->file_size = (long)finfo.st_size;
 	}
       free (filename);
-      return;
+      return i;
     }
 
-  mailfiles = (FILEINFO **)
-    xrealloc (mailfiles,
-	      ((++mailfiles_count) * sizeof (FILEINFO *)));
+  i = mailfiles_count++;
+  mailfiles = (FILEINFO **)xrealloc
+		(mailfiles, mailfiles_count * sizeof (FILEINFO *));
 
-  mailfiles[mailfiles_count - 1] = (FILEINFO *)xmalloc (sizeof (FILEINFO));
-  mailfiles[mailfiles_count - 1]->name = filename;
+  mailfiles[i] = (FILEINFO *)xmalloc (sizeof (FILEINFO));
+  mailfiles[i]->name = filename;
   if (stat (filename, &finfo) == 0)
     {
-      mailfiles[mailfiles_count - 1]->access_time = finfo.st_atime;
-      mailfiles[mailfiles_count - 1]->mod_time = finfo.st_mtime;
-      mailfiles[mailfiles_count - 1]->file_size = finfo.st_size;
+      mailfiles[i]->access_time = finfo.st_atime;
+      mailfiles[i]->mod_time = finfo.st_mtime;
+      mailfiles[i]->file_size = finfo.st_size;
     }
   else
     {
-      mailfiles[mailfiles_count - 1]->access_time
-	= mailfiles[mailfiles_count - 1]->mod_time = (time_t)-1;
-      mailfiles[mailfiles_count - 1]->file_size = (long)-1;
+      mailfiles[i]->access_time = mailfiles[i]->mod_time = (time_t)-1;
+      mailfiles[i]->file_size = -1L;
     }
+  return i;
 }
 
 /* Reset the existing mail files access and modification times to zero. */
+void
 reset_mail_files ()
 {
   register int i;
@@ -150,17 +151,20 @@ reset_mail_files ()
   for (i = 0; i < mailfiles_count; i++)
     {
       mailfiles[i]->access_time = mailfiles[i]->mod_time = 0;
-      mailfiles[i]->file_size = (long)0;
+      mailfiles[i]->file_size = 0L;
     }
 }
 
 /* Free the information that we have about the remembered mail files. */
+void
 free_mail_files ()
 {
-  while (mailfiles_count--)
+  register int i;
+
+  for (i = 0; i < mailfiles_count; i++)
     {
-      free (mailfiles[mailfiles_count]->name);
-      free (mailfiles[mailfiles_count]);
+      free (mailfiles[i]->name);
+      free (mailfiles[i]);
     }
 
   if (mailfiles)
@@ -170,70 +174,71 @@ free_mail_files ()
   mailfiles = (FILEINFO **)NULL;
 }
 
-/* Return non-zero if FILE's mod date has changed. */
+/* Return non-zero if FILE's mod date has changed and it has not been
+   accessed since modified. */
+static int
 file_mod_date_changed (file)
      char *file;
 {
-  time_t time = (time_t)NULL;
+  time_t time = (time_t)0;
   struct stat finfo;
+  int i;
 
-  int index = find_mail_file (file);
-  if (index != -1)
-    time = mailfiles[index]->mod_time;
+  i = find_mail_file (file);
+  if (i != -1)
+    time = mailfiles[i]->mod_time;
+
+  if ((stat (file, &finfo) == 0) && (finfo.st_size > 0))
+    return (time != finfo.st_mtime);
 
-  if (stat (file, &finfo) == 0)
-    {
-      if (finfo.st_size != 0)
-	return (time != finfo.st_mtime);
-    }
   return (0);
 }
 
 /* Return non-zero if FILE's access date has changed. */
+static int
 file_access_date_changed (file)
      char *file;
 {
-  time_t time = (time_t)NULL;
+  time_t time = (time_t)0;
   struct stat finfo;
+  int i;
 
-  int index = find_mail_file (file);
-  if (index != -1)
-    time = mailfiles[index]->access_time;
+  i = find_mail_file (file);
+  if (i != -1)
+    time = mailfiles[i]->access_time;
+
+  if ((stat (file, &finfo) == 0) && (finfo.st_size > 0))
+    return (time != finfo.st_atime);
 
-  if (stat (file, &finfo) == 0)
-    {
-      if (finfo.st_size != 0)
-	return (time != finfo.st_atime);
-    }
   return (0);
 }
 
 /* Return non-zero if FILE's size has increased. */
+static int
 file_has_grown (file)
      char *file;
 {
   long size = 0L;
   struct stat finfo;
+  int i;
 
-  int i = find_mail_file (file);
+  i = find_mail_file (file);
   if (i != -1)
     size = mailfiles[i]->file_size;
 
-  if (stat (file, &finfo) == 0)
-    {
-      return (finfo.st_size > size);
-    }
-  return (0);
+  return ((stat (file, &finfo) == 0) && (finfo.st_size > size));
 }
 
 #if defined (USG)
-#define DEFAULT_MAIL_PATH "/usr/mail/"
+#  define DEFAULT_MAIL_PATH "/usr/mail/"
+#  define DEFAULT_PATH_LEN 10
 #else
-#define DEFAULT_MAIL_PATH "/usr/spool/mail/"
+#  define DEFAULT_MAIL_PATH "/usr/spool/mail/"
+#  define DEFAULT_PATH_LEN 16
 #endif
 
 /* Return the colon separated list of pathnames to check for mail. */
-char *
+static char *
 get_mailpaths ()
 {
   char *mailpaths;
@@ -243,15 +248,13 @@ get_mailpaths ()
   if (!mailpaths)
     mailpaths = get_string_value ("MAIL");
 
-  if (!mailpaths)
-    {
-      mailpaths = (char *)
-	alloca (1 + strlen (DEFAULT_MAIL_PATH) + strlen (current_user.user_name));
-
-      sprintf (mailpaths, "%s%s", DEFAULT_MAIL_PATH, current_user.user_name);
-    }
-
-  return (savestring (mailpaths));
+  if (mailpaths)
+    return (savestring (mailpaths));
+    
+  mailpaths = xmalloc (1 + DEFAULT_PATH_LEN + strlen (current_user.user_name));
+  strcpy (mailpaths, DEFAULT_MAIL_PATH);
+  strcpy (mailpaths + DEFAULT_PATH_LEN, current_user.user_name);
+  return (mailpaths);
 }
 
 /* Take an element from $MAILPATH and return the portion from
@@ -285,13 +288,14 @@ parse_mailpath_spec (str)
 /* Remember the dates of the files specified by MAILPATH, or if there is
    no MAILPATH, by the file specified in MAIL.  If neither exists, use a
    default value, which we randomly concoct from using Unix. */
+void
 remember_mail_dates ()
 {
   char *mailpaths = get_mailpaths ();
   char *mailfile, *mp;
-  int index = 0, pass_next = 0;
+  int i = 0, pass_next = 0;
   
-  while (mailfile = extract_colon_unit (mailpaths, &index))
+  while (mailfile = extract_colon_unit (mailpaths, &i))
     {
       mp = parse_mailpath_spec (mailfile);
       if (mp && *mp)
@@ -314,12 +318,13 @@ remember_mail_dates ()
    Special hack:  If the shell variable MAIL_WARNING is on and the
    mail file has been accessed since the last time we remembered, then
    the message "The mail in <mailfile> has been read" is printed. */
+void
 check_mail ()
 {
   register int string_index;
   char *current_mail_file, *you_have_mail_message;
-  char *mailpaths = get_mailpaths (), *mp;
-  int index = 0;
+  char *mailpaths, *mp;
+  int file_index = 0;
   char *dollar_underscore;
 
   dollar_underscore = get_string_value ("_");
@@ -327,7 +332,8 @@ check_mail ()
   if (dollar_underscore)
     dollar_underscore = savestring (dollar_underscore);
 
-  while ((current_mail_file = extract_colon_unit (mailpaths, &index)))
+  mailpaths = get_mailpaths ();
+  while ((current_mail_file = extract_colon_unit (mailpaths, &file_index)))
     {
       char *t;
       int use_user_notification;
@@ -358,14 +364,17 @@ check_mail ()
 	  WORD_LIST *tlist;
 	  int i, file_is_bigger;
 	  bind_variable ("_", current_mail_file);
+#define atime mailfiles[i]->access_time
+#define mtime mailfiles[i]->mod_time
 
 	  /* Have to compute this before the call to add_mail_file, which
 	     resets all the information. */
 	  file_is_bigger = file_has_grown (current_mail_file);
 
-	  add_mail_file (current_mail_file);
+	  i = add_mail_file (current_mail_file);
 
-	  i = find_mail_file (current_mail_file);
+	  if (i == -1)
+	    continue;		/* if this returns -1 , it is a bug */
 
 	  /* If the user has just run a program which manipulates the
 	     mail file, then don't bother explaining that the mail
@@ -373,46 +382,37 @@ check_mail ()
 	     the access time to be equal to the modification time when
 	     the mail in the file is manipulated, check the size also.  If
 	     the file has not grown, continue. */
-	  if (i != -1 &&
-	      (mailfiles[i]->access_time >= mailfiles[i]->mod_time) &&
-	      !file_is_bigger)
-	    goto next_mail_file;
-
-#if 0
-	  /* XXX - this test is useless, and so is the assignment.
-		   have already gone to the next mail file if the test
-		   succeeds, since it is the same as the one above, and
-		   the assignment has already been made. */
-	  if (!use_user_notification)
+	  if ((atime >= mtime) && !file_is_bigger)
 	    {
-	      if (i != -1 &&
-		  (mailfiles[i]->access_time < mailfiles[i]->mod_time) &&
-		  file_is_bigger)
-		you_have_mail_message = "You have new mail in $_";
+	      free (current_mail_file);
+	      continue;
 	    }
-#endif
+
+	  /* If the mod time is later than the access time and the file
+	     has grown, note the fact that this is *new* mail. */
+	  if (!use_user_notification && (atime < mtime) && file_is_bigger)
+	    you_have_mail_message = "You have new mail in $_";
+#undef atime
+#undef mtime
 
 	  if ((tlist = expand_string (you_have_mail_message, 1)))
 	    {
 	      char *tem = string_list (tlist);
-	      you_have_mail_message = (char *)alloca (1 + strlen (tem));
-	      strcpy (you_have_mail_message, tem);
+	      printf ("%s\n", tem);
 	      free (tem);
+	      dispose_words (tlist);
 	    }
 	  else
-	    you_have_mail_message = "";
-
-	  printf ("%s\n", you_have_mail_message);
-	  dispose_words (tlist);
+	    printf ("\n");
 	}
 
-      if (find_variable ("MAIL_WARNING")
-	  && file_access_date_changed (current_mail_file))
+      if (find_variable ("MAIL_WARNING") &&
+	    file_access_date_changed (current_mail_file))
 	{
 	  add_mail_file (current_mail_file);
 	  printf ("The mail in %s has been read!\n", current_mail_file);
 	}
-    next_mail_file: 
+
       free (current_mail_file);
     }
   free (mailpaths);
