@@ -20,17 +20,19 @@ with Bash; see the file COPYING.  If not, write to the Free Software
 Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include <stdio.h>
-#include <sys/types.h>
+#include "bashtypes.h"
 #include <sys/file.h>
+#include "filecntl.h"
 #include "bashansi.h"
 #include "config.h"
+#include "command.h"
 #include "general.h"
 #include "error.h"
-#include "command.h"
 #include "flags.h"
-#include "filecntl.h"
 #include "make_cmd.h"
 #include "subst.h"
+#include "input.h"
+#include "externs.h"
 
 #if defined (JOB_CONTROL)
 #include "jobs.h"
@@ -39,9 +41,6 @@ Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 extern int line_number, current_command_line_count;
 extern int disallow_filename_globbing;
 
-extern char *redirection_expand ();
-extern char *read_secondary_line ();
-
 WORD_DESC *
 make_word (string)
      char *string;
@@ -119,6 +118,7 @@ add_string_to_list (string, list)
   return (temp);
 }
 
+#if 0
 WORD_DESC *
 coerce_to_word (number)
      int number;
@@ -128,6 +128,7 @@ coerce_to_word (number)
   sprintf (string, "%d", number);
   return (make_word (string));
 }
+#endif
 
 COMMAND *
 make_command (type, pointer)
@@ -174,6 +175,23 @@ make_for_command (name, map_list, action)
   return (make_command (cm_for, (SIMPLE_COM *)temp));
 }
 
+#if defined (SELECT_COMMAND)
+COMMAND *
+make_select_command (name, map_list, action)
+     WORD_DESC *name;
+     WORD_LIST *map_list;
+     COMMAND *action;
+{
+  SELECT_COM *temp = (SELECT_COM *)xmalloc (sizeof (SELECT_COM));
+
+  temp->flags = 0;
+  temp->name = name;
+  temp->map_list = map_list;
+  temp->action = action;
+  return (make_command (cm_select, (SIMPLE_COM *)temp));
+}
+#endif
+
 COMMAND *
 make_group_command (command)
      COMMAND *command;
@@ -206,7 +224,7 @@ make_pattern_list (patterns, action)
   PATTERN_LIST *temp;
 
   temp = (PATTERN_LIST *)xmalloc (sizeof (PATTERN_LIST));
-  temp->patterns = patterns;
+  temp->patterns = REVERSE_LIST (patterns, WORD_LIST *);
   temp->action = action;
   temp->next = NULL;
   return (temp);
@@ -226,7 +244,7 @@ make_if_command (test, true_case, false_case)
   return (make_command (cm_if, (SIMPLE_COM *)temp));
 }
 
-COMMAND *
+static COMMAND *
 make_until_or_while (test, action, which)
      COMMAND *test, *action;
      enum command_type which;
@@ -323,12 +341,12 @@ make_here_document (temp)
 
       case r_deblank_reading_until:	/* <<-foo */
 	kill_leading++;
-	/* ... */
+	/* FALLTHROUGH */
       case r_reading_until:		/* <<foo */
 	{
-	  char *redirectee_word;
-	  int len;
-
+	  char *redir_word;
+	  int redir_len;
+	  char *full_line;
 	  char *document = (char *)NULL;
 	  int document_index = 0, document_size = 0;
 
@@ -339,7 +357,7 @@ make_here_document (temp)
 	    int old_value = disallow_filename_globbing;
 	    disallow_filename_globbing = 1;
 
-	    redirectee_word = redirection_expand (temp->redirectee.filename);
+	    redir_word = redirection_expand (temp->redirectee.filename);
 
 	    disallow_filename_globbing = old_value;
 	  }
@@ -347,15 +365,14 @@ make_here_document (temp)
 	  /* Quote removal is the only expansion performed on the delimiter
 	     for here documents, making it an extremely special case.  I
 	     still feel ill. */
-	  redirectee_word =
-	    string_quote_removal (temp->redirectee.filename->word, 0);
+	  redir_word = string_quote_removal (temp->redirectee.filename->word, 0);
 #endif /* POSIX_HERE_DOCUMENTS */
 
 	  /* redirection_expand will return NULL if the expansion results in
 	     multiple words or no words.  Check for that here, and just abort
 	     this here document if it does. */
-	  if (redirectee_word)
-	    len = strlen (redirectee_word);
+	  if (redir_word)
+	    redir_len = strlen (redir_word);
 	  else
 	    {
 	      temp->here_doc_eof = savestring ("");
@@ -363,73 +380,66 @@ make_here_document (temp)
 	    }
 
 	  free (temp->redirectee.filename->word);
-	  temp->here_doc_eof = redirectee_word;
+	  temp->here_doc_eof = redir_word;
 
 	  /* Read lines from wherever lines are coming from.
 	     For each line read, if kill_leading, then kill the
 	     leading tab characters.
-	     If the line matches redirectee_word exactly, then we have
+	     If the line matches redir_word exactly, then we have
 	     manufactured the document.  Otherwise, add the line to the
 	     list of lines in the document. */
-	  {
-	    char *line;
-	    int l;
 
-	    /* If the here-document delimiter was quoted, the lines should
-	       be read verbatim from the input.  If it was not quoted, we
-	       need to perform backslash-quoted newline removal. */
-	    while (line = read_secondary_line
-		     (temp->redirectee.filename->quoted == 0))
-	      {
-		line_number++;
+	  /* If the here-document delimiter was quoted, the lines should
+	     be read verbatim from the input.  If it was not quoted, we
+	     need to perform backslash-quoted newline removal. */
+	  while (full_line = read_secondary_line
+		 (temp->redirectee.filename->quoted == 0))
+	    {
+	      register char *line = full_line;
+	      int len;
 
-		if (kill_leading)
-		  {
-		    register int i;
+	      line_number++;
 
-		    /* Hack:  To be compatible with some Bourne shells, we 
-		       check the word before stripping the whitespace.  This
-		       is a hack, though. */
-		    if ((strncmp (line, redirectee_word, len) == 0) &&
-			line[len] == '\n')
-		      {
-			free (line);
-			goto document_done;
-		      }
-
-		    for (i = 0; line[i] == '\t'; i++)
-		      ;
-
-		    if (i)
-		      strcpy (&line[0], &line[i]);
-		  }
-
-		if ((strncmp (line, redirectee_word, len) == 0) &&
-		    line[len] == '\n')
-		  {
-		    free (line);
+	      if (kill_leading && *line)
+	        {
+		  /* Hack:  To be compatible with some Bourne shells, we 
+		     check the word before stripping the whitespace.  This
+		     is a hack, though. */
+		  if (STREQN (line, redir_word, redir_len) &&
+		      line[redir_len] == '\n')
 		    goto document_done;
-		  }
 
-		l = strlen (line);
-		if (l + document_index >= document_size)
-		  {
-		    document = (char *)
-		      xrealloc (document, (document_size += (10 * l)));
-		  }
+		  while (*line == '\t')
+		    line++;
+		}
+
+	      if (!*line)
+	        continue;
+
+	      if (STREQN (line, redir_word, redir_len) &&
+		  line[redir_len] == '\n')
+		goto document_done;
+
+	      len = strlen (line);
+	      if (len + document_index >= document_size)
+		{
+		  document_size = document_size ? 2 * (document_size + len)
+						: 1000;	/* XXX */
+		  document = xrealloc (document, document_size);
+		}
+
+	      /* len is guaranteed to be > 0 because of the check for line
+		 being an empty string before the call to strlen. */
+	      FASTCOPY (line, document + document_index, len);
+	      document_index += len;
+	    }
 
-		if (l != 0)
-		  {
-		    strcpy (&document[document_index], line);
-		    document_index += l;
-		  }
-		free (line);
-	      }
   document_done:
-	    if (!document)
-	      document = savestring ("");
-	    temp->redirectee.filename->word = document;
-	  }
+	  if (document)
+	    document[document_index] = '\0';
+	  else
+	    document = savestring ("");
+	  temp->redirectee.filename->word = document;
 	}
     }
 }
@@ -441,13 +451,13 @@ REDIRECT *
 make_redirection (source, instruction, dest_and_filename)
      int source;
      enum r_instruction instruction;
-     WORD_DESC *dest_and_filename;
+     REDIRECTEE dest_and_filename;
 {
   REDIRECT *temp = (REDIRECT *)xmalloc (sizeof (REDIRECT));
 
   /* First do the common cases. */
   temp->redirector = source;
-  temp->redirectee.filename = dest_and_filename;
+  temp->redirectee = dest_and_filename;
   temp->instruction = instruction;
   temp->flags = 0;
   temp->next = (REDIRECT *)NULL;
@@ -547,10 +557,46 @@ make_word_array (list)
 
   for (count = 0; list; count++)
     {
-      array[count] = (char *)xmalloc (1 + strlen (list->word->word));
+      array[count] = xmalloc (1 + strlen (list->word->word));
       strcpy (array[count], list->word->word);
       list = list->next;
     }
   array[count] = (char *)NULL;
   return (array);
 }
+
+/* The Yacc grammar productions have a problem, in that they take a
+   list followed by an ampersand (`&') and do a simple command connection,
+   making the entire list effectively asynchronous, instead of just
+   the last command.  This means that when the list is executed, all
+   the commands have stdin set to /dev/null when job control is not
+   active, instead of just the last.  This is wrong, and needs fixing
+   up.  This function takes the `&' and applies it to the last command
+   in the list. */
+COMMAND *
+connect_async_list (command, command2, connector)
+     COMMAND *command, *command2;
+     int connector;
+{
+  COMMAND *t, *t1, *t2;
+
+  t1 = command;
+  t = command->value.Connection->second;
+  /* This is just defensive programming.  The Yacc precedence rules
+     will generally hand this function a command where t points directly
+     to the command we want (e.g. given a ; b ; c ; d &, t1 will point
+     to the `a ; b ; c' list and t will be the `d').  We only want to do
+     this if the list is not being executed as a unit in the background
+     with `( ... )', so we have to check for CMD_WANT_SUBSHELL.  That's
+     the only way to tell. */
+  while (((t->flags & CMD_WANT_SUBSHELL) == 0) && (t->type == cm_connection))
+    {
+      t1 = t;
+      t = t->value.Connection->second;
+    }
+  /* Now we have t pointing to the last command in the list, and
+     t1->value.Connection->second == t. */
+  t2 = command_connect (t, command2, connector);
+  t1->value.Connection->second = t2;
+  return command;
+}
