@@ -21,15 +21,17 @@
    Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include <stdio.h>
-#include <sys/types.h>
+#include "bashtypes.h"
 #include <signal.h>
 #include <setjmp.h>
 #include <errno.h>
 
 #include "config.h"
+#include "command.h"
 #include "general.h"
 #include "filecntl.h"
 #include "jobs.h"
+#include "externs.h"
 
 #if defined (BUFFERED_INPUT)
 #  include "input.h"
@@ -227,11 +229,9 @@ sigwinch_sighandler (sig)
 #endif /* !READLINE && TIOCGWINSZ && SIGWINCH */
 
 /* Setup this shell to handle C-C, etc. */
+void
 initialize_job_signals ()
 {
-  extern int login_shell;
-  extern sighandler sigint_sighandler ();
-
   set_signal_handler (SIGINT, sigint_sighandler);
 #if !defined (READLINE) && defined (TIOCGWINSZ) && defined (SIGWINCH)
   set_signal_handler (SIGWINCH, sigwinch_sighandler);
@@ -313,12 +313,12 @@ make_child (command, async_p)
 	}
 #endif /* _POSIX_VERSION */
 
-      report_error ("fork: %s", strerror (errno));
+      internal_error ("fork: %s", strerror (errno));
 
       throw_to_top_level ();
     }
  
-  if (!pid)
+  if (pid == 0)
     {
 #if defined (BUFFERED_INPUT)
       if (default_buffered_input > 0)
@@ -336,8 +336,11 @@ make_child (command, async_p)
       /* Ignore INT and QUIT in asynchronous children. */
       if (async_p)
 	{
+#if 0
+	  /* This now done by setup_async_signals (). */
 	  set_signal_handler (SIGINT, SIG_IGN);
 	  set_signal_handler (SIGQUIT, SIG_IGN);
+#endif
 	  last_asynchronous_pid = getpid ();
 	}
 
@@ -407,28 +410,26 @@ wait_for_single_pid (pid)
 void
 wait_for_background_pids ()
 {
+  pid_t got_pid;
+  WAIT status;
+
   /* If we aren't using job control, we let the kernel take care of the
      bookkeeping for us.  wait () will return -1 and set errno to ECHILD 
      when there are no more unwaited-for child processes on both
      4.2 BSD-based and System V-based systems. */
 
   siginterrupt (SIGINT, 1);
-  while (1)
+
+  /* Wait for ECHILD */
+  while ((got_pid = WAITPID (-1, &status, 0)) != -1)
+    set_pid_status (got_pid, status);
+
+  if (errno != EINTR && errno != ECHILD)
     {
-      pid_t got_pid;
-      WAIT status;
-
-      /* Wait for ECHILD */
-      while ((got_pid = WAITPID (-1, &status, 0)) != -1)
-	set_pid_status (got_pid, status);
-
-      if (errno != EINTR && errno != ECHILD)
-	{
-	  siginterrupt (SIGINT, 0);
-	  file_error("wait");
-	}
-      break;
+      siginterrupt (SIGINT, 0);
+      file_error("wait");
     }
+
   siginterrupt (SIGINT, 0);
   QUIT;
 }
@@ -456,7 +457,6 @@ int
 wait_for (pid)
      pid_t pid;
 {
-  extern int interactive, interactive_shell, last_command_exit_value;
   int return_val, pstatus;
   pid_t got_pid;
   WAIT status;
@@ -475,7 +475,7 @@ wait_for (pid)
   if (!interactive_shell)
     old_sigint_handler = set_signal_handler (SIGINT, wait_sigint_handler);
 
-  while ((got_pid = WAITPID (pid, &status, 0)) != pid)
+  while ((got_pid = WAITPID (-1, &status, 0)) != pid) /* XXX was pid now -1 */
     {
       if (got_pid < 0 && errno == ECHILD)
 	{
@@ -521,7 +521,7 @@ wait_for (pid)
   if (!WIFSTOPPED (status) && WIFSIGNALED (status) &&
       (WTERMSIG (status) != SIGINT))
     {
-      fprintf (stderr, "%s", sys_siglist[WTERMSIG (status)]);
+      fprintf (stderr, "%s", strsignal (WTERMSIG (status)));
       if (WIFCORED (status))
 	fprintf (stderr, " (core dumped)");
       fprintf (stderr, "\n");
@@ -622,19 +622,21 @@ give_terminal_to (pgrp)
 /* Stop a pipeline. */
 stop_pipeline (async, ignore)
      int async;
-     char *ignore;
+     COMMAND *ignore;
 {
   already_making_children = 0;
 }
 
+void
 start_pipeline ()
 {
   already_making_children = 1;
 }
 
 /* Print descriptive information about the job with leader pid PID. */
+void
 describe_pid (pid)
      pid_t pid;
 {
-  fprintf (stderr, "<%d>\n", (int) pid);
+  fprintf (stderr, "%d\n", (int) pid);
 }
