@@ -32,14 +32,14 @@
 */
 #define INSTALL_DEBUG_MODE
 
-#include <sys/types.h>
+#include "bashtypes.h"
 #include <stdio.h>
 #include <signal.h>
 #include <errno.h>
 #include <sys/file.h>
+#include "filecntl.h"
 #include <pwd.h>
 #include "posixstat.h"
-#include "filecntl.h"
 #include "bashansi.h"
 
 #if defined (HAVE_VARARGS_H)
@@ -53,40 +53,41 @@
 #include "jobs.h"
 #endif /* JOB_CONTROL */
 
-#if defined (BUFFERED_INPUT)
-#  include "input.h"
-#endif /* BUFFERED_INPUT */
-
-#if defined (USG) && !defined (isc386) && !defined (sgi) && !defined (Linux)
-extern struct passwd *getpwuid ();
-#endif
-
-extern char *dist_version;
-extern int build_version;
+#include "input.h"
+#include "execute_cmd.h"
 
 #if defined (HISTORY)
-extern void using_history ();
-#endif /* HISTORY */
+#  include "bashhist.h"
+#  include <readline/history.h>
+#endif
 
-extern char *find_path_file ();
-extern char *tilde_expand ();
-extern char *decode_prompt_string ();
+#include <tilde/tilde.h>
+
+#if defined (USG) && !defined (HAVE_GETPW_DECLS)
+extern struct passwd *getpwuid ();
+#endif /* USG && !HAVE_GETPW_DECLS */
 
 extern int yydebug;
 #if !defined (errno)
 extern int errno;
 #endif
 
+extern char *dist_version;
+extern int patch_level, build_version;
 extern int subshell_environment; /* Found in execute_cmd.c. */
-
-#if defined (RESTRICTED_SHELL)
-/* Non-zero means this shell is restricted. */
-extern int restricted;		/* in flags.c */
-#endif
+extern int last_command_exit_value;
+extern int return_catch_flag;
+extern jmp_buf return_catch;
+extern int need_here_doc, current_command_line_count, line_number;
+extern char *ps1_prompt, **prompt_string_pointer;
+extern int loop_level, continuing, breaking;
+extern int parse_and_execute_level;
+extern char *this_command_name;
 
 /* Non-zero means that this shell has already been run; i.e. you should
    call shell_reinitialize () if you need to start afresh. */
 static int shell_initialized = 0;
+static int sourced_env = 0;
 
 /* The current maintainer of the shell.  You change this in the
    Makefile. */
@@ -96,14 +97,7 @@ static int shell_initialized = 0;
 
 char *the_current_maintainer = MAINTAINER;
 
-#if !defined (PPROMPT)
-#  define PPROMPT "bash\\$ "
-#endif
 char *primary_prompt = PPROMPT;
-
-#if !defined (SPROMPT)
-#  define SPROMPT "> "
-#endif
 char *secondary_prompt = SPROMPT;
 
 COMMAND *global_command = (COMMAND *)NULL;
@@ -143,13 +137,6 @@ int interactive_shell = 0;
 */
 int startup_state = 0;
 
-#if defined (HISTORY)
-/* Non-zero means to remember lines typed to the shell on the history
-   list.  This is different than the user-controlled behaviour; this
-   becomes zero when we read lines from a file, for example. */
-int remember_on_history = 1;
-#endif /* HISTORY */
-
 /* Special debugging helper. */
 int debugging_login_shell = 0;
 
@@ -225,14 +212,6 @@ struct {
   { (char *)0x0, Int, (int *)0x0, (char **)0x0 }
 };
 
-#if defined (HISTORY)
-/* The number of lines that Bash has added to this history session. */
-int history_lines_this_session = 0;
-
-/* The number of lines that Bash has read from the history file. */
-int history_lines_in_file = 0;
-#endif /* HISTORY */
-
 /* These are extern so execute_simple_command can set them, and then
    longjmp back to main to execute a shell script, instead of calling
    main () again and resulting in indefinite, possibly fatal, stack
@@ -247,34 +226,29 @@ char **subshell_envp;
 int default_buffered_input = -1;
 #endif
 
-static char *local_pending_command = (char *)NULL;
+static int want_pending_command;
+static char *local_pending_command;
 
 static int isnetconn ();
 static void run_startup_files ();
 
+static void shell_initialize ();
+static void shell_reinitialize ();
+static void initialize_signals ();
+static void initialize_terminating_signals ();
+
 main (argc, argv, env)
      int argc;
      char **argv, **env;
 {
-  extern int last_command_exit_value;
   register int i;
   int arg_index, locally_skip_execution;
   int top_level_arg_index, read_from_stdin;
   FILE *default_input;
-#if defined (JOB_CONTROL)
-  extern int job_control;
-#endif /* JOB_CONTROL */
-
-#if defined (AUX)
-#include <compat.h>
-  set42sig ();
-  setcompat (getcompat() | COMPAT_BSDGROUPS | COMPAT_BSDSIGNALS |
-	     COMPAT_BSDTTY | COMPAT_EXEC | COMPAT_SYSCALLS);
-#endif /* AUX */
 
   /* There is a bug in the NeXT 2.1 rlogind that causes opens
      of /dev/tty to fail. */
-#if defined (NeXT)
+#if defined (RLOGIN_PGRP_BUG)
   {
     int tty_fd;
 
@@ -288,7 +262,7 @@ main (argc, argv, env)
       }
     close (tty_fd);
   }
-#endif /* NeXT */
+#endif /* RLOGIN_PGRP_BUG */
 
   /* Wait forever if we are debugging a login shell. */
   while (debugging_login_shell);
@@ -306,11 +280,7 @@ main (argc, argv, env)
 		    (getenv ("POSIX_PEDANTIC") != (char *)NULL);
 
 #if defined (USE_GNU_MALLOC_LIBRARY)
-  {
-    extern void programming_error ();
-
-    mcheck (programming_error, (void (*) ())0);
-  }
+  mcheck (programming_error, (void (*) ())0);
 #endif /* USE_GNU_MALLOC_LIBRARY */
 
   if (setjmp (subshell_top_level))
@@ -318,11 +288,13 @@ main (argc, argv, env)
       argc = subshell_argc;
       argv = subshell_argv;
       env = subshell_envp;
+      sourced_env = 0;
     }
 
   /* Initialize local variables for all `invocations' of main (). */
   arg_index = 1;
   local_pending_command = (char *)NULL;
+  want_pending_command = 0;
   locally_skip_execution = 0;
   read_from_stdin = 0;
   default_input = stdin;
@@ -349,15 +321,12 @@ main (argc, argv, env)
 
   /* Here's a hack.  If the name of this shell is "sh", then don't do
      any startup files; just try to be more like /bin/sh. */
-  {
-    char *tshell_name = base_pathname (argv[0]);
-
-    if (*tshell_name == '-')
-      tshell_name++;
-
-    if (strcmp (tshell_name, "sh") == 0)
-      act_like_sh++;
-  }
+  /* XXX - next version - make this be the same as -posix. */
+  shell_name = base_pathname (argv[0]);
+  if (*shell_name == '-')
+    shell_name++;
+  if (shell_name[0] == 's' && shell_name[1] == 'h' && !shell_name[2])
+    act_like_sh++;
 
   yydebug = 0;
 
@@ -373,7 +342,7 @@ main (argc, argv, env)
 
 #if defined (JOB_CONTROL)
   if (act_like_sh)
-    job_control = 0;
+    job_control = 0;		/* XXX - not posix */
 #endif /* JOB_CONTROL */
 
   shell_start_time = NOW;	/* NOW now defined in general.h */
@@ -381,7 +350,7 @@ main (argc, argv, env)
   /* A program may start an interactive shell with
 	  "execl ("/bin/bash", "-", NULL)".
      If so, default the name of this shell to our name. */
-  if (!shell_name || !*shell_name || (strcmp (shell_name, "-") == 0))
+  if (!shell_name || !*shell_name || (shell_name[0] == '-' && !shell_name[1]))
     shell_name = "bash";
 
   /* Parse argument flags from the input line. */
@@ -391,16 +360,16 @@ main (argc, argv, env)
     {
       for (i = 0; long_args[i].name; i++)
 	{
-	  if (strcmp (&(argv[arg_index][1]), long_args[i].name) == 0)
+	  if (STREQ (&(argv[arg_index][1]), long_args[i].name))
 	    {
 	      if (long_args[i].type == Int)
-		*(long_args[i].int_value) = 1;
+		*long_args[i].int_value = 1;
 	      else
 		{
 		  if (!argv[++arg_index])
 		    {
-		      report_error ("%s: Flag `%s' expected an argument",
-				    shell_name, long_args[i].name);
+		      report_error ("option `%s' expected an argument",
+				    long_args[i].name);
 		      exit (1);
 		    }
 		  else
@@ -425,62 +394,57 @@ main (argc, argv, env)
       login_shell = -login_shell;
     }
 
-  /* All done with full word args; do standard shell arg parsing.*/
+  /* All done with full word options; do standard shell option parsing.*/
+  this_command_name = shell_name;	/* for error reporting */
   while (arg_index != argc && argv[arg_index] &&
-	 (*(argv[arg_index]) == '-' || (*argv[arg_index] == '+')))
+	 (*argv[arg_index] == '-' || *argv[arg_index] == '+'))
     {
       /* There are flag arguments, so parse them. */
-      int arg_character;
-      int on_or_off = (*argv[arg_index]);
-      char *o_option;
-      int next_arg = arg_index + 1;
+      int arg_character, on_or_off, next_arg;
+      char *o_option, *arg_string;
 
       i = 1;
+      next_arg = arg_index + 1;
+      arg_string = argv[arg_index];
+      on_or_off = arg_string[0];
 
       /* A single `-' signals the end of options.  From the 4.3 BSD sh.
 	 An option `--' means the same thing; this is the standard
-	 getopt () meaning. */
-      if (((argv[arg_index][0] == '-') && (argv[arg_index][1] == '\0')) ||
-	  (strcmp (argv[arg_index], "--") == 0))
+	 getopt(3) meaning. */
+      if (arg_string[0] == '-' &&
+	   (arg_string[1] == '\0' ||
+	     (arg_string[1] == '-' && arg_string[2] == '\0')))
 	{
 	  arg_index++;
-	  goto after_flags;
+	  break;
 	}
 
-      while (arg_character = (argv[arg_index])[i++])
+      while (arg_character = arg_string[i++])
 	{
 	  switch (arg_character)
 	    {
 	    case 'c':
-	      /* The next arg is a command to execute, and the following args
-		 are $1 .. $n respectively. */
-		local_pending_command = argv[++arg_index];
-		if (!local_pending_command)
-		  {
-		    report_error ("`%cc' requires an argument", on_or_off);
-		    exit (1);
-		  }
-		arg_index++;
-		goto after_flags;
+		want_pending_command = 1;
+		break;
 
 	    case 's':
 		read_from_stdin = 1;
 		break;
 
 	    case 'o':
-		o_option = argv[next_arg++];
+		o_option = argv[next_arg];
 		if (!o_option)
 		  {
-		    report_error ("`%co' requires an argument", on_or_off);
-		    exit (1);
+		    list_minus_o_opts ();
+		    break;
 		  }
-
 		if (set_minus_o_option (on_or_off, o_option) != EXECUTION_SUCCESS)
 		  exit (1);
+		next_arg++;
 		break;
 
 	    default:
-	      if (change_flag_char (arg_character, on_or_off) == FLAG_ERROR)
+	      if (change_flag (arg_character, on_or_off) == FLAG_ERROR)
 		{
 		  report_error ("%c%c: bad option", on_or_off, arg_character);
 		  exit (1);
@@ -489,11 +453,24 @@ main (argc, argv, env)
 	    }
 	}
       /* Can't do just a simple increment anymore -- what about
-	 "bash -abouo emacs ignoreeof -hO"? */
+	 "bash -abouo emacs ignoreeof -hP"? */
       arg_index = next_arg;
     }
 
- after_flags:
+  /* Need to get the argument to a -c option processed in the
+     above loop.  The next arg is a command to execute, and the
+     following args are $0...$n respectively. */
+  if (want_pending_command)
+    {
+      local_pending_command = argv[arg_index];
+      if (!local_pending_command)
+	{
+	  report_error ("`-c' requires an argument");
+	  exit (1);
+	}
+      arg_index++;
+    }
+  this_command_name = (char *)NULL;      
 
   /* First, let the outside world know about our interactive status.
      A shell is interactive if the `-i' flag was given, or if all of
@@ -511,17 +488,14 @@ main (argc, argv, env)
        isatty (fileno (stdin)) &&	/* Input is a terminal and */
        isatty (fileno (stdout))))	/* output is a terminal. */
     {
-      interactive_shell = startup_state = 1;
-      interactive = 1;
+      interactive_shell = startup_state = interactive = 1;
     }
   else
     {
 #if defined (HISTORY)
-      history_expansion = 0;
-      remember_on_history = 0;
+      history_expansion = remember_on_history = 0;
 #endif /* HISTORY */
-      interactive_shell = startup_state = 0;
-      interactive = 0;
+      interactive_shell = startup_state = interactive = 0;
       no_line_editing = 1;
 #if defined (JOB_CONTROL)
       job_control = 0;
@@ -529,7 +503,6 @@ main (argc, argv, env)
     }
 
 #define CLOSE_FDS_AT_LOGIN
-
 #if defined (CLOSE_FDS_AT_LOGIN)
   /*
    * Some systems have the bad habit of starting login shells with lots of open
@@ -552,8 +525,7 @@ main (argc, argv, env)
   if (interactive_shell)
     {
       char *term = getenv ("TERM");
-      if (term && (strcmp (term, "emacs") == 0))
-	no_line_editing = 1;
+      no_line_editing = term && (STREQ (term, "emacs"));
     }
 
   top_level_arg_index = arg_index;
@@ -590,7 +562,7 @@ main (argc, argv, env)
     }
   else
     {
-      change_flag_char ('i', FLAG_ON);
+      change_flag ('i', FLAG_ON);
       interactive = 1;
     }
 
@@ -606,24 +578,19 @@ main (argc, argv, env)
 
     if (local_pending_command)
       {
-	/* Bind remaining args to $1 ... $n */
+	/* Bind remaining args to $0 ... $n */
 	WORD_LIST *args = (WORD_LIST *)NULL;
 	while (arg_index != argc)
 	  args = make_word_list (make_word (argv[arg_index++]), args);
 	if (args)
 	  {
-	    args = (WORD_LIST *)reverse_list (args);
-	    if (posixly_correct)
-	      {
-		/* Posix.2 4.56.3 says that the first argument after
-		   sh -c command becomes $0, and the rest of the arguments
-		   are bound to $1 ... $N. */
-		shell_name = savestring (args->word->word);
-		dollar_vars[0] = savestring (shell_name);
-		remember_args (args->next, 1);
-	      }
-	    else
-	      remember_args (args, 1);
+	    args = REVERSE_LIST (args, WORD_LIST *);
+	    /* Posix.2 4.56.3 says that the first argument after
+	       sh -c command becomes $0, and the rest of the arguments
+	       are bound to $1 ... $N. */
+	    shell_name = savestring (args->word->word);	/* XXX */
+	    dollar_vars[0] = savestring (args->word->word);
+	    remember_args (args->next, 1);
 	    dispose_words (args);
 	  }
 
@@ -641,31 +608,9 @@ main (argc, argv, env)
   if (interactive_shell)
     {
       /* Set up for checking for presence of mail. */
-#if defined (USG)
-      /* Under System V, we can only tell if you have mail if the
-	 modification date has changed.  So remember the current
-	 modification dates. */
       remember_mail_dates ();
-#else
-      /* Under 4.x, you have mail if there is something in your inbox.
-	 I set the remembered mail dates to 1900.  */
-      reset_mail_files ();
-#endif /* USG */
-
-      /* If this was a login shell, then assume that /bin/login has already
-	 taken care of informing the user that they have new mail.  Otherwise,
-	 we want to check right away. */
-      if (login_shell == 1)
-	{
-#if !defined (USG)
-	  remember_mail_dates ();
-#endif
-	}
-
       reset_mail_timer ();
 
-      change_flag_char ('i', FLAG_ON);
-
 #if defined (HISTORY)
       /* Initialize the interactive history stuff. */
       if (!shell_initialized)
@@ -688,7 +633,7 @@ main (argc, argv, env)
       filename = savestring (argv[arg_index]);
 
       fd = open (filename, O_RDONLY);
-      if ((fd < 0) && (errno == ENOENT))
+      if ((fd < 0) && (errno == ENOENT) && (absolute_program (filename) == 0))
 	{
 	  char *path_filename;
 	  /* If it's not in the current directory, try looking through PATH
@@ -808,7 +753,7 @@ main (argc, argv, env)
     WORD_LIST *args = (WORD_LIST *)NULL;
     while (arg_index != argc)
       args = make_word_list (make_word (argv[arg_index++]), args);
-    args = (WORD_LIST *)reverse_list (args);
+    args = REVERSE_LIST (args, WORD_LIST *);
     remember_args (args, 1);
     dispose_words (args);
   }
@@ -880,26 +825,24 @@ main (argc, argv, env)
 static void
 run_startup_files ()
 {
-  static int sourced_env = 0;
-
   if (!posixly_correct)
     {
       if (login_shell)
 	{
 	  /* We don't execute .bashrc for login shells. */
           no_rc++;
-          maybe_execute_file (SYS_PROFILE);
+          maybe_execute_file (SYS_PROFILE, 1);
         }
 
       if (login_shell && !no_profile)
         {
           if (act_like_sh)
-            maybe_execute_file ("~/.profile");
+            maybe_execute_file ("~/.profile", 1);
           else
             {
-              if (maybe_execute_file ("~/.bash_profile") == 0)
-	        if (maybe_execute_file ("~/.bash_login") == 0)
-	          maybe_execute_file ("~/.profile");
+              if (maybe_execute_file ("~/.bash_profile", 1) == 0)
+	        if (maybe_execute_file ("~/.bash_login", 1) == 0)
+	          maybe_execute_file ("~/.profile", 1);
             }
         }
 
@@ -927,7 +870,7 @@ run_startup_files ()
       if (!act_like_sh && !no_rc &&
           (interactive_shell || (isnetconn (fileno (stdin)) &&
 			         local_pending_command)))
-	maybe_execute_file (bashrc_file);
+        maybe_execute_file (bashrc_file, 1);
     }
 
    /* Try a TMB suggestion.  If running a script, then execute the
@@ -954,7 +897,7 @@ run_startup_files ()
 	      dispose_words (list);
 
 	      if (expanded_file_name && *expanded_file_name)
-		maybe_execute_file (expanded_file_name);
+		maybe_execute_file (expanded_file_name, 1);
 
 	      if (expanded_file_name)
 		free (expanded_file_name);
@@ -975,7 +918,7 @@ maybe_make_restricted (name)
   char *temp;
 
   temp = base_pathname (shell_name);
-  if (restricted || (strcmp (temp, "rbash") == 0))
+  if (restricted || (STREQ (temp, "rbash"))
     {
       set_var_read_only ("PATH");
       non_unsettable ("PATH");
@@ -986,108 +929,18 @@ maybe_make_restricted (name)
 }
 #endif /* RESTRICTED_SHELL */
 
-#if defined (HISTORY)
-/* If this is an interactive shell, then append the lines executed
-   this session to the history file. */
-int
-maybe_save_shell_history ()
-{
-  int result = 0;
-
-  if (interactive && history_lines_this_session)
-    {
-      void using_history ();
-      char *hf = get_string_value ("HISTFILE");
-
-      if (hf && *hf)
-	{
-	  struct stat buf;
-
-	  /* If the file doesn't exist, then create it. */
-	  if (stat (hf, &buf) == -1)
-	    {
-	      int file = open (hf, O_CREAT | O_TRUNC | O_WRONLY, 0666);
-	      if (file != -1)
-		close (file);
-	    }
-
-	  /* Now actually append the lines if the history hasn't been
-	     stifled. */
-	  using_history ();
-	  if (history_lines_this_session <= where_history ())
-	    {
-	      result = append_history (history_lines_this_session, hf);
-	      history_lines_in_file += history_lines_this_session;
-	      history_lines_this_session = 0;
-	    }
-	}
-    }
-  return (result);
-}
-
-/* Load the history list from the history file. */
-load_history ()
-{
-  char *hf;
-
-  /* Truncate history file for interactive shells which desire it.
-     Note that the history file is automatically truncated to the
-     size of HISTSIZE if the user does not explicitly set the size
-     differently. */
-  set_if_not ("HISTFILESIZE", get_string_value ("HISTSIZE"));
-  stupidly_hack_special_variables ("HISTFILESIZE");
-
-  /* Read the history in HISTFILE into the history list. */
-  hf = get_string_value ("HISTFILE");
-
-  if (hf && *hf)
-    {
-      struct stat buf;
-
-      if (stat (hf, &buf) == 0)
-	{
-	  read_history (hf);
-	  using_history ();
-	  history_lines_in_file = where_history ();
-	}
-    }
-}
-
-/* Write the existing history out to the history file. */
-save_history ()
-{
-  char *hf = get_string_value ("HISTFILE");
-
-  if (hf && *hf)
-    {
-      struct stat buf;
-
-      if (stat (hf, &buf) == 0)
-	{
-	  /* Append only the lines that occurred this session to
-	     the history file. */
-	  using_history ();
-
-	  if (history_lines_this_session < where_history ())
-	    append_history (history_lines_this_session, hf);
-	  else
-	    write_history (hf);
-	}
-    }
-}
-#endif /* HISTORY */
-
 /* Try to execute the contents of FNAME.  If FNAME doesn't exist,
-   that is not an error, but other kinds of errors are.  Returns
-   -1 in the case of an error, 0 in the case that the file was not
-   found, and 1 if the file was found and executed. */
-maybe_execute_file (fname)
+   that is not an error, but other kinds of errors are.  A non-zero
+   FORCE_NONINTERACTIVE means to set the value of `interactive' to
+   0 so things like job control are disabled; the value is unchanged
+   otherwise.  Returns -1 in the case of an error, 0 in the case that
+   the file was not found, and 1 if the file was found and executed. */
+maybe_execute_file (fname, force_noninteractive)
      char *fname;
+     int force_noninteractive;
 {
-  extern int return_catch_flag;
-  extern jmp_buf return_catch;
   jmp_buf old_return_catch;
-  int return_val, fd, tresult;
+  int return_val, fd, tresult, old_interactive;
   char *filename, *string;
   struct stat file_info;
 
@@ -1123,7 +976,13 @@ file_error_and_exit:
   string[file_info.st_size] = '\0';
 
   return_catch_flag++;
-  bcopy ((char *)return_catch, (char *)old_return_catch, sizeof (jmp_buf));
+  xbcopy ((char *)return_catch, (char *)old_return_catch, sizeof (jmp_buf));
+
+  if (force_noninteractive)
+    {
+      old_interactive = interactive;
+      interactive = 0;
+    }
 
   return_val = setjmp (return_catch);
 
@@ -1134,8 +993,11 @@ file_error_and_exit:
   else
     tresult = parse_and_execute (string, filename);
 
+  if (force_noninteractive)
+    interactive = old_interactive;
+
   return_catch_flag--;
-  bcopy ((char *)old_return_catch, (char *)return_catch, sizeof (jmp_buf));
+  xbcopy ((char *)old_return_catch, (char *)return_catch, sizeof (jmp_buf));
 
   free (filename);
 
@@ -1148,7 +1010,6 @@ file_error_and_exit:
 run_one_command (command)
      char *command;
 {
-  sighandler sigint_sighandler ();
   int code;
 
   code = setjmp (top_level);
@@ -1182,8 +1043,6 @@ reader_loop ()
 
   while (!EOF_Reached)
     {
-      extern char *trap_list[];
-      sighandler sigint_sighandler ();
       int code;
 
       code = setjmp (top_level);
@@ -1307,13 +1166,23 @@ alrm_catcher(i)
 
 parse_command ()
 {
-  extern int need_here_doc, current_command_line_count;
   int r;
 
   need_here_doc = 0;
-
   run_pending_traps ();
 
+  /* Allow the execution of a random command just before the printing
+     of each primary prompt.  If the shell variable PROMPT_COMMAND
+     is set then the value of it is the command to execute. */
+  if (interactive)
+    {
+      char *command_to_execute;
+
+      command_to_execute = get_string_value ("PROMPT_COMMAND");
+      if (command_to_execute)
+	execute_prompt_command (command_to_execute);
+    }
+
   current_command_line_count = 0;
   r = yyparse ();
 
@@ -1325,8 +1194,6 @@ parse_command ()
 
 read_command ()
 {
-  extern char *ps1_prompt, **prompt_string_pointer;
-  extern int current_command_line_count;
   SHELL_VAR *tmout_var = (SHELL_VAR *)NULL;
   int tmout_len = 0, result;
   SigHandler *old_alrm = (SigHandler *)NULL;
@@ -1386,6 +1253,7 @@ line_buffer_stream (stream)
 
 /* Do whatever is necessary to initialize the shell.
    Put new initializations in here. */
+static void
 shell_initialize ()
 {
   /* Line buffer output for stderr and stdout. */
@@ -1399,7 +1267,7 @@ shell_initialize ()
   /* Initialize the trap signal handlers before installing our own
      signal handlers.  traps.c:restore_original_signals () is responsible
      for restoring the original default signal handlers.  That function
-     is called from jobs.c when we make a new child. */
+     is called when we make a new child. */
   initialize_traps ();
   initialize_signals ();
 
@@ -1449,12 +1317,11 @@ shell_initialize ()
 }
 
 /* Function called by main () when it appears that the shell has already
-   had some initialization preformed.  This is supposed to reset the world
+   had some initialization performed.  This is supposed to reset the world
    back to a pristine state, as if we had been exec'ed. */
+static void
 shell_reinitialize ()
 {
-  extern int line_number, last_command_exit_value;
-
   /* The default shell prompts. */
   primary_prompt = PPROMPT;
   secondary_prompt = SPROMPT;
@@ -1463,15 +1330,15 @@ shell_reinitialize ()
   current_command_number = 1;
 
   /* We have decided that the ~/.bashrc file should not be executed
-     for the invocation of each shell script.  Perhaps some other file
-     should.  */
-  act_like_sh = 1;
+     for the invocation of each shell script.  If the variable $ENV
+     (or $BASH_ENV) is set, its value is used as the name of a file
+     to source. */
+  no_rc = no_profile = 1;
 
   /* Things that get 0. */
   login_shell = make_login_shell = interactive = executing = 0;
-  debugging = no_rc = no_profile = do_version = line_number = 0;
+  debugging = do_version = line_number = last_command_exit_value = 0;
   forced_interactive = interactive_shell = subshell_environment = 0;
-  last_command_exit_value = 0;
 
 #if defined (HISTORY)
   remember_on_history = history_expansion = 0;
@@ -1495,6 +1362,7 @@ shell_reinitialize ()
   sv_path ("PATH");
 }
 
+static void
 initialize_signals ()
 {
   initialize_terminating_signals ();
@@ -1623,11 +1491,7 @@ termination_unwind_protect (sig)
 
 #if defined (JOB_CONTROL)
   if (interactive && login_shell && sig == SIGHUP)
-    {
-      extern void hangup_all_jobs ();
-
-      hangup_all_jobs ();
-    }
+    hangup_all_jobs ();
   end_job_control ();
 #endif /* JOB_CONTROL */
 
@@ -1646,6 +1510,7 @@ termination_unwind_protect (sig)
 
 /* Initialize signals that will terminate the shell to do some
    unwind protection. */
+static void
 initialize_terminating_signals ()
 {
   register int i;
@@ -1660,19 +1525,36 @@ initialize_terminating_signals ()
   act.sa_handler = termination_unwind_protect;
   act.sa_flags = 0;
   sigemptyset (&act.sa_mask);
+  sigemptyset (&oact.sa_mask);
   for (i = 0; i < TERMSIGS_LENGTH; i++)
     sigaddset (&act.sa_mask, XSIG (i));
   for (i = 0; i < TERMSIGS_LENGTH; i++)
     {
       sigaction (XSIG (i), &act, &oact);
       terminating_signals[i].orig_handler = oact.sa_handler;
+      /* Don't do anything with signals that are ignored at shell entry
+	 if the shell is not interactive. */
+      if (!interactive_shell && oact.sa_handler == SIG_IGN)
+        {
+	  sigaction (XSIG (i), &oact, &act);
+	  set_signal_ignored (XSIG (i));
+        }
     }
 
 #else /* !_POSIX_VERSION */
 
   for (i = 0; i < TERMSIGS_LENGTH; i++)
-    terminating_signals[i].orig_handler =
-      signal (XSIG (i), termination_unwind_protect);
+    {
+      terminating_signals[i].orig_handler =
+	signal (XSIG (i), termination_unwind_protect);
+      /* Don't do anything with signals that are ignored at shell entry
+	 if the shell is not interactive. */
+      if (!interactive_shell && terminating_signals[i].orig_handler == SIG_IGN)
+	{
+          signal (XSIG (i), SIG_IGN);
+          set_signal_ignored (XSIG (i));
+	}
+    }
 
 #endif /* !_POSIX_VERSION */
 
@@ -1689,8 +1571,6 @@ initialize_terminating_signals ()
 
   if (interactive)
     {
-      extern sighandler sigint_sighandler ();
-
       set_signal_handler (SIGINT, sigint_sighandler);
       signal (SIGTERM, SIG_IGN);
     }
@@ -1715,7 +1595,6 @@ reset_terminating_signals ()
 
       act.sa_handler = XHANDLER (i);
       sigaction (XSIG (i), &act, (struct sigaction *) NULL);
-#undef XSIG
     }
 #else
   for (i = 0; i < TERMSIGS_LENGTH; i++)
@@ -1734,9 +1613,6 @@ reset_terminating_signals ()
 void
 throw_to_top_level ()
 {
-  extern int last_command_exit_value, loop_level, continuing, breaking;
-  extern int return_catch_flag;
-  extern int parse_and_execute_level;
   int print_newline = 0;
 
   if (interrupt_state)
@@ -1788,7 +1664,8 @@ throw_to_top_level ()
     }
 
   /* An interrupted `wait' command in a script does not exit the script. */
-  if (interactive || (print_newline && signal_is_trapped (SIGINT)))
+  if (interactive || (interactive_shell && !shell_initialized) ||
+      (print_newline && signal_is_trapped (SIGINT)))
     longjmp (top_level, DISCARD);
   else
     longjmp (top_level, EXITPROG);
@@ -1821,129 +1698,31 @@ sigint_sighandler (sig)
 #endif /* VOID_SIGHANDLER */
 }
 
-#if defined (MAKE_BUG_REPORTS)
-/* Make a bug report, even to the extent of mailing it.  Hope that it
-   gets where it is supposed to go.  If not, maybe the user will send
-   it back to me. */
-#if defined (HISTORY)
-#  include <readline/history.h>
-
-  /* Number of commands to place in the bug report. */
-#  define LAST_INTERESTING_HISTORY_COUNT 6
-#endif /* HISTORY */
-
-#if defined (HAVE_VFPRINTF)
-make_bug_report (va_alist)
-     va_dcl
-#else
-make_bug_report (reason, arg1, arg2)
-     char *reason;
-#endif /* HAVE_VFPRINTF */
-{
-  extern char *current_host_name, *the_current_maintainer;
-  extern int interactive, login_shell;
-  FILE *stream, *popen ();
-#if defined (HISTORY)
-  register int len = where_history ();
-  register int i = len - LAST_INTERESTING_HISTORY_COUNT;
-  HIST_ENTRY **list = history_list ();
-#endif /* HISTORY */
-
-#if defined (HAVE_VFPRINTF)
-  char *reason;
-  va_list args;
-#endif /* HAVE_VFPRINTF */
-
-  stream = popen ("/bin/rmail bash-maintainers@ai.mit.edu", "w");
-
-#if defined (HISTORY)
-  save_history ();
-  if (i < 0) i = 0;
-#endif /* HISTORY */
-
-  if (stream)
-    {
-      fprintf (stream, "To: bash-maintainers@ai.mit.edu\n");
-      fprintf (stream, "Subject: Bash-%s.%d bug-report: ",
-	       dist_version, build_version);
-
-#if defined (HAVE_VFPRINTF)
-      va_start (args);
-      reason = va_arg (args, char *);
-      vfprintf (stream, reason, args);
-      va_end (args);
-#else
-      fprintf (stream, reason, arg1, arg2);
-#endif /* HAVE_VFPRINTF */
-
-      fprintf (stream, "\n");
-
-      /* Write the history into the mail file.  Maybe we can recreate
-	 the bug? */
-      fprintf (stream,
-	       "This is a Bash bug report.  Bash maintainers should be getting this report.\n\
-If this mail has bounced, for right now please send it to:\n\
-\n\
-	%s\n\
-\n\
-since he is the current maintainer of this version of the shell.\n\
-\n\
-This is %s (invoked as `%s'), version %s.%d, on host %s, used by %s.\n\
-This shell is %sinteractive, and it is %sa login shell.\n\
-\n\
-The host is a %s running %s.\n\
-\n\
-The current environment is:\n",
-	       the_current_maintainer,
-	       get_string_value ("BASH"), full_pathname (dollar_vars[0]),
-	       dist_version, build_version, current_host_name,
-	       current_user.user_name, interactive ? "" : "not ",
-	       login_shell ? "" : "not ", SYSTEM_NAME, OS_NAME);
-
-      {
-	SHELL_VAR **vlist, *var;
-	register int i;
-
-	vlist = all_shell_variables ();
-
-	for (i = 0; vlist && var = vlist[i]; i++)
-	  {
-	    if (!invisible_p (var) && exported_p (var))
-	      {
-		fprintf (stream, "%s=%s", var->name, value_cell (var));
-		fprintf (stream, "\n");
-	      }
-	  }
-      }
-
-#if defined (HISTORY)
-      fprintf (stream, "\nAnd here are the last %d commands.\n\n",
-	       LAST_INTERESTING_HISTORY_COUNT);
-
-      for (; i < len; i++)
-	fprintf (stream, "%s\n", list[i]->line);
-#endif /* HISTORY */
-
-      pclose (stream);
-    }
-  else
-    {
-      fprintf (stderr, "Can't mail bug report!\n");
-    }
-}
-#endif /* MAKE_BUG_REPORTS */
-
 /* Give version information about this shell. */
+char *
+shell_version_string ()
+{
+  static char tt[16] = { '\0' };
+
+  if (!tt[0])
+    sprintf (tt, "%s.%d(%d)", dist_version, patch_level, build_version);
+  return tt;
+}
+
+void
 show_shell_version ()
 {
-  extern char *shell_name;
-  extern int version;
-
-  printf ("GNU %s, version %s.%d-CWRU\n", base_pathname (shell_name),
-	  dist_version, build_version);
+  printf ("GNU %s, version %s\n", base_pathname (shell_name),
+    shell_version_string ());
 }
 
 #if !defined (USG) && defined (ENOTSOCK)
+#  if !defined (HAVE_SOCKETS)
+#    define HAVE_SOCKETS
+#  endif
+#endif
+
+#if defined (HAVE_SOCKETS)
 #include <sys/socket.h>
 #endif
 
@@ -1960,17 +1739,14 @@ isnetconn (fd)
     return (0);
   return (S_ISCHR (sb.st_mode));
 #else /* !USGr4 && !USGr4_2 */
-#  if defined (ENOTSOCK) && !defined (USG)
+#  if defined (HAVE_SOCKETS)
   int rv, l;
   struct sockaddr sa;
-                  
+
   l = sizeof(sa);
   rv = getpeername(0, &sa, &l);
-  if (rv < 0 && errno == ENOTSOCK)
-    return 0;
-  else
-    return 1;
-#  else /* !ENOTSOCK && !USG */
+  return ((rv < 0 && errno == ENOTSOCK) ? 0 : 1);
+#  else /* !HAVE_SOCKETS */
 #    if defined (S_ISSOCK)
   struct stat sb;
 
@@ -1980,6 +1756,6 @@ isnetconn (fd)
 #    else /* !S_ISSOCK */
   return (0);
 #    endif /* !S_ISSOCK */
-#  endif /* !ENOTSOCK && !USG */
+#  endif /* !HAVE_SOCKETS */
 #endif /* !USGr4 && !USGr4_2 */
 }
