@@ -25,9 +25,12 @@
 /* #define STANDALONE */
 
 #include <stdio.h>
-#include <sys/types.h>
+#include "bashtypes.h"
 
 #if !defined (STANDALONE)
+#  if !defined (_POSIX_VERSION)
+#    include <sys/file.h>
+#  endif /* !_POSIX_VERSION */
 #  include "posixstat.h"
 #  include "filecntl.h"
 #  include "shell.h"
@@ -39,11 +42,14 @@
 #  if defined (HAVE_UNISTD_H)
 #    include <unistd.h>
 #  endif /* HAVE_UNISTD_H */
+#  define whitespace(c) (((c) == ' ') || ((c) == '\t'))
+#  define digit(c)  ((c) >= '0' && (c) <= '9')
+#  define digit_value(c) ((c) - '0')
 #endif /* STANDALONE */
 
-#if !defined (_POSIX_VERSION)
-#  include <sys/file.h>
-#endif /* !_POSIX_VERSION */
+#if !defined (STRLEN)
+#  define STRLEN(s) ((s)[0] ? ((s)[1] ? ((s)[2] ? strlen(s) : 2) : 1) : 0)
+#endif
 
 #include <errno.h>
 #if !defined (errno)
@@ -69,22 +75,23 @@ extern int errno;
 #endif /* !_POSIX_VERSION */
 
 /* What type are the user and group ids?  GID_T is actually the type of
-   the array that getgroups(3) returns. */
-#if defined (SunOS4) || defined (__bsdi__)
+   the members of the array that getgroups(3) fills in from its second
+   argument. */
+#if defined (INT_GROUPS_ARRAY)
 #  define GID_T int
 #  define UID_T int
-#else /* !SunOS4 && !__bsdi__*/
+#else /* !INT_GROUPS_ARRAY */
 #  define GID_T gid_t
 #  define UID_T uid_t
-#endif /* !SunOS4 && !__bsdi__ */
+#endif /* !INT_GROUPS_ARRAY */
 
-#if !defined (Linux) && !defined (USGr4_2)
+#if !defined (Linux) && !defined (USGr4_2) && !defined (SunOS5)
 extern gid_t getegid ();
 extern uid_t geteuid ();
 #  if !defined (sony)
 extern gid_t getgid ();
 #  endif /* !sony */
-#endif /* !Linux && !USGr4_2 */
+#endif /* !Linux && !USGr4_2 && !SunOS5 */
 
 #if !defined (R_OK)
 #define R_OK 4
@@ -115,10 +122,20 @@ extern gid_t getgid ();
 	do { test_error_return = val; longjmp (test_exit_buf, 1); } while (0)
 #endif /* STANDALONE */
 
+#if defined (AFS)
+  /* We have to use access(2) for machines running AFS, because it's
+     not a Unix file system.  This may produce incorrect answers for
+     non-AFS files.  I hate AFS. */
+#  define EACCESS(path, mode)	access(path, mode)
+#else
+#  define EACCESS(path, mode)	eaccess(path, mode)
+#endif /* AFS */
+
 static int pos;		/* The offset of the current argument in ARGV. */
 static int argc;	/* The number of arguments present in ARGV. */
 static char **argv;	/* The argument list. */
 
+static int isint ();
 static int unop ();
 static int binop ();
 static int unary_operator ();
@@ -136,13 +153,20 @@ static void
 test_syntax_error (format, arg)
      char *format, *arg;
 {
+#if !defined (STANDALONE)
+  extern int interactive_shell;
+  extern char *get_name_for_error ();
+  if (!interactive_shell)
+    fprintf (stderr, "%s: ", get_name_for_error ());
+#endif
   fprintf (stderr, "%s: ", argv[0]);
   fprintf (stderr, format, arg);
   fflush (stderr);
   test_exit (SHELL_BOOLEAN (FALSE));
 }
 
-/* A wrapper for stat () which disallows pathnames that are empty strings. */
+/* A wrapper for stat () which disallows pathnames that are empty strings
+   and handles /dev/fd emulation on systems that don't have it. */
 static int
 test_stat (path, finfo)
      char *path;
@@ -153,6 +177,19 @@ test_stat (path, finfo)
       errno = ENOENT;
       return (-1);
     }
+#if !defined (HAVE_DEV_FD)
+  if (path[0] == '/' && path[1] == 'd' && strncmp (path, "/dev/fd/", 8) == 0)
+    {
+      int fd;
+      if (isint (path + 9, &fd))
+	return (fstat (fd, finfo));
+      else
+	{
+	  errno = EBADF;
+	  return (-1);
+	}
+    }
+#endif /* !HAVE_DEV_FD */
   return (stat (path, finfo));
 }
 
@@ -164,7 +201,6 @@ eaccess (path, mode)
      char *path;
      int mode;
 {
-  extern int group_member ();
   struct stat st;
   static int euid = -1;
 
@@ -217,17 +253,17 @@ int
 group_member (gid)
      GID_T gid;
 {
-  static GID_T pgid = NOGROUP;
-  static GID_T egid = NOGROUP;
+  static GID_T pgid = (GID_T)NOGROUP;
+  static GID_T egid = (GID_T)NOGROUP;
 
-  if (pgid == NOGROUP)
+  if (pgid == (GID_T)NOGROUP)
 #if defined (SHELL)
     pgid = (GID_T) current_user.gid;
 #else /* !SHELL */
     pgid = (GID_T) getgid ();
 #endif /* !SHELL */
 
-  if (egid == NOGROUP)
+  if (egid == (GID_T)NOGROUP)
 #if defined (SHELL)
     egid = (GID_T) current_user.egid;
 #else /* !SHELL */
@@ -469,7 +505,7 @@ binary_operator ()
   /* Are the left and right integer expressions of the form '-l string'? */
   int l_is_l, r_is_l;
 
-  if (strcmp (argv[pos], "-l") == 0)
+  if (argv[pos][0] == '-' && argv[pos][1] == 'l' && !argv[pos][2])
     {
       l_is_l = 1;
       op = pos + 2;
@@ -486,7 +522,8 @@ binary_operator ()
       op = pos + 1;
     }
 
-  if ((op < argc - 2) && (strcmp (argv[op + 1], "-l") == 0))
+  if ((op < argc - 2) &&
+      (argv[op + 1][0] == '-' && argv[op + 1][1] == 'l' && !argv[op + 1][2]))
     {
       r_is_l = 1;
       advance (0);
@@ -682,14 +719,16 @@ binary_operator ()
 
   if (argv[op][0] == '=' && !argv[op][1])
     {
-      value = (strcmp (argv[pos], argv[pos + 2]) == 0);
+      value = (argv[pos][0] == argv[pos+2][0]) &&
+	      (strcmp (argv[pos], argv[pos + 2]) == 0);
       pos += 3;
       return (TRUE == value);
     }
 
-  if (strcmp (argv[op], "!=") == 0)
+  if (argv[op][0] == '!' && argv[op][1] == '=' && !argv[op][2])
     {
-      value = (strcmp (argv[pos], argv[pos + 2]) != 0);
+      value = (argv[pos][0] != argv[pos + 2][0]) ||
+	      (strcmp (argv[pos], argv[pos + 2]) != 0);
       pos += 3;
       return (TRUE == value);
     }
@@ -720,21 +759,17 @@ unary_operator ()
 
     case 'r':			/* file is readable? */
       unary_advance ();
-      value = -1 != eaccess (argv[pos - 1], R_OK);
+      value = -1 != EACCESS (argv[pos - 1], R_OK);
       return (TRUE == value);
 
     case 'w':			/* File is writeable? */
       unary_advance ();
-      value = -1 != eaccess (argv[pos - 1], W_OK);
+      value = -1 != EACCESS (argv[pos - 1], W_OK);
       return (TRUE == value);
 
     case 'x':			/* File is executable? */
       unary_advance ();
-#if defined (AFS)
-      value = -1 != access (argv[pos - 1], X_OK);
-#else
-      value = -1 != eaccess (argv[pos - 1], X_OK);
-#endif
+      value = -1 != EACCESS (argv[pos - 1], X_OK);
       return (TRUE == value);
 
     case 'O':			/* File is owned by you? */
@@ -762,8 +797,12 @@ unary_operator ()
 
       /* Under POSIX, -f is true if the given file exists
 	 and is a regular file. */
+#if defined (S_IFMT)
       return (TRUE == ((S_ISREG (stat_buf.st_mode)) ||
 		       (0 == (stat_buf.st_mode & S_IFMT))));
+#else
+      return (TRUE == (S_ISREG (stat_buf.st_mode)));
+#endif /* !S_IFMT */
 
     case 'd':			/* File is a directory? */
       unary_advance ();
@@ -886,7 +925,7 @@ and ()
   int value;
 
   value = term ();
-  while ((pos < argc) && strcmp (argv[pos], "-a") == 0)
+  while (pos < argc && argv[pos][0] == '-' && argv[pos][1] == 'a' && !argv[pos][2])
     {
       advance (0);
       value = TRUTH_AND (value, and ());
@@ -906,7 +945,7 @@ or ()
 
   value = and ();
 
-  while ((pos < argc) && strcmp (argv[pos], "-o") == 0)
+  while (pos < argc && argv[pos][0] == '-' && argv[pos][1] == 'o' && !argv[pos][2])
     {
       advance (0);
       value = TRUTH_OR (value, or ());
@@ -952,8 +991,8 @@ two_arguments ()
 {
   int value;
 
-  if (STREQ (argv[pos], "!"))
-    value = strlen (argv[pos + 1]) == 0;
+  if (argv[pos][0] == '!' && !argv[pos][1])
+    value = argv[pos + 1][0] == '\0';
   else if ((argv[pos][0] == '-') && (argv[pos][2] == '\0'))
     {
       if (unop (argv[pos][1]))
@@ -962,7 +1001,8 @@ two_arguments ()
 	test_syntax_error ("%s: unary operator expected\n", argv[pos]);
     }
   else
-    beyond ();
+    test_syntax_error ("%s: unary operator expected\n", argv[pos]);
+
   return (value);
 }
 
@@ -971,7 +1011,7 @@ three_arguments ()
 {
   int value;
 
-  if (STREQ (argv[pos], "!"))
+  if (argv[pos][0] == '!' && !argv[pos][1])
     {
       advance (1);
       value = !two_arguments ();
@@ -981,7 +1021,9 @@ three_arguments ()
       value = binary_operator ();
       pos = argc;
     }
-  else if ((STREQ (argv[pos+1], "-a")) || (STREQ (argv[pos+1], "-o")) ||
+  /* Check for -a or -o or a parenthesized subexpression. */
+  else if ((argv[pos+1][0] == '-' && !argv[pos+1][2] &&
+  		(argv[pos+1][1] == 'a' || argv[pos+1][1] == 'o')) ||
 	   (argv[pos][0] == '('))
     value = expr ();
   else
@@ -1003,7 +1045,7 @@ posixtest ()
 	break;
 
       case 1:
-	value = strlen (argv[1]) != 0;
+	value = argv[1][0] != '\0';
 	pos = argc;
 	break;
 
@@ -1060,14 +1102,14 @@ test_command (margc, margv)
 
   argv = margv;
 
-  if (margv[0] && strcmp (margv[0], "[") == 0)
+  if (margv[0] && margv[0][0] == '[' && !margv[0][1])
     {
       --margc;
 
       if (margc < 2)
 	test_exit (SHELL_BOOLEAN (FALSE));
 
-      if (margv[margc] && strcmp (margv[margc], "]") != 0)
+      if (margv[margc] && (margv[margc][0] != ']' || margv[margc][1]))
 	test_syntax_error ("missing `]'\n", (char *)NULL);
     }
 
