@@ -19,7 +19,8 @@
    Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include <stdio.h>
-#include <sys/types.h>
+#include "bashtypes.h"
+#include "posixstat.h"
 #include <ctype.h>
 #include <pwd.h>
 
@@ -27,15 +28,22 @@
 #include "shell.h"
 #include "hash.h"
 #include "flags.h"
+#include "execute_cmd.h"
 
 #include "builtins/common.h"
+#include <tilde/tilde.h>
 
-extern char *indirection_level_string ();
-extern char *find_user_command ();
-extern char *tilde_expand (), *named_function_string ();
-
+/* Variables used here and defined in other files. */
 extern int posixly_correct;
-extern struct user_info current_user;
+extern int variable_context, line_number;
+extern int interactive_shell, login_shell, shell_level;
+extern int subshell_environment;
+extern int build_version;
+extern char *dist_version;
+extern char *shell_name;
+extern char *primary_prompt, *secondary_prompt;
+extern Function *this_shell_builtin;
+extern time_t shell_start_time;
 
 /* The list of shell variables that the user has created, or that came from
    the environment. */
@@ -91,27 +99,22 @@ static char *have_local_variables;		/* XXX */
 static int local_variable_stack_size = 0;	/* XXX */
 
 /* Some forward declarations. */
-SHELL_VAR *set_if_not ();	/* returns new or existing entry */
-static void sbrand ();		/* set bash random number generator */
-void non_unsettable ();
+static void initialize_dynamic_variables ();
+static void sbrand ();		/* set bash random number generator. */
 
 /* Make VAR be auto-exported.  VAR is a pointer to a SHELL_VAR. */
 #define set_auto_export(var) \
 { var->attributes |= att_exported; array_needs_making = 1; }
 
 #if defined (HISTORY)
-extern int remember_on_history;
+#  include "bashhist.h"
 #endif /* HISTORY */
 
 /* Initialize the shell variables from the current environment. */
+void
 initialize_shell_variables (env)
-     char *env[];
+     char **env;
 {
-  extern char *primary_prompt, *secondary_prompt;
-  extern int interactive_shell;
-#if defined (HISTORY)
-  extern int remember_on_history;
-#endif /* HISTORY */
   char *name, *string, *current_dir;
   int c, char_index;
   int string_index = 0;
@@ -130,7 +133,7 @@ initialize_shell_variables (env)
       char_index = 0;
 
       string_length = strlen (string);
-      name = (char *)alloca (1 + string_length);
+      name = xmalloc (1 + string_length);
 
       while ((c = *string++) && c != '=')
 	name[char_index++] = c;
@@ -143,7 +146,7 @@ initialize_shell_variables (env)
 	  SHELL_VAR *f;
 	  char *eval_string;
 
-	  eval_string = (char *)xmalloc (3 + string_length + strlen (name));
+	  eval_string = xmalloc (3 + string_length + strlen (name));
 	  sprintf (eval_string, "%s %s", name, string);
 	  parse_and_execute (eval_string, name);
 
@@ -166,6 +169,7 @@ initialize_shell_variables (env)
 	  v->attributes |= (att_exported | att_imported);
 	  array_needs_making = 1;
 	}
+      free (name);
     }
 
   /* If we got PWD from the environment, update our idea of the current
@@ -173,28 +177,11 @@ initialize_shell_variables (env)
      checking it.  It is possible for getwd () to fail on shell startup,
      and in that case, PWD would be undefined. */
   temp_var = find_variable ("PWD");
-  if (temp_var && imported_p (temp_var))
-    {
-      char *curdir;
-
-      if (curdir = value_cell (temp_var))
-	{
-	  if (same_file
-	      (curdir, ".", (struct stat *)NULL, (struct stat *)NULL))
-	    set_working_directory (curdir);
-	  else
-	    {
-	      current_dir = get_working_directory ("shell-init");
-	      if (current_dir)
-		{
-		  bind_variable ("PWD", current_dir);
-		  free (current_dir);
-		}
-	    }
-	}
-    }
+  if (temp_var && imported_p (temp_var) &&
+      (current_dir = value_cell (temp_var)) &&
+      same_file (current_dir, ".", (struct stat *)NULL, (struct stat *)NULL))
+    set_working_directory (current_dir);
   else
-    /* Set up $PWD. */
     {
       current_dir = get_working_directory ("shell-init");
       if (current_dir)
@@ -230,6 +217,8 @@ initialize_shell_variables (env)
   /* Magic machine types.  Pretty convenient. */
   temp_var = set_if_not ("HOSTTYPE", HOSTTYPE);
   set_auto_export (temp_var);
+  temp_var = set_if_not ("OSTYPE", OSTYPE);
+  set_auto_export (temp_var);
 
   /* Default MAILPATH, and MAILCHECK for interactive shells. */
   if (interactive_shell)
@@ -241,8 +230,8 @@ initialize_shell_variables (env)
 	{
 	  char *tem;
 
-	  tem = (char *)xmalloc (1 + sizeof (DEFAULT_MAIL_PATH)
-				 + strlen (current_user.user_name));
+	  tem = xmalloc (1 + sizeof (DEFAULT_MAIL_PATH)
+			   + strlen (current_user.user_name));
 	  strcpy (tem, DEFAULT_MAIL_PATH);
 	  strcat (tem, current_user.user_name);
 
@@ -280,19 +269,9 @@ initialize_shell_variables (env)
   bind_variable ("OPTERR", "1");
 #endif /* GETOPTS_BUILTIN */
 
-  /* Find out if we're supposed to be in Posix.2 mode via an
-     environment variable. */
-  temp_var = find_variable ("POSIXLY_CORRECT");
-  if (!temp_var)
-    temp_var = find_variable ("POSIX_PEDANTIC");
-  if (temp_var && imported_p (temp_var))
-    sv_strict_posix (temp_var->name);
-
   /* Get the full pathname to THIS shell, and set the BASH variable
      to it. */
   {
-    extern char *shell_name;
-    extern int login_shell;
     char *tname = find_user_command (shell_name);
 
     if ((login_shell == 1) && (*shell_name != '/'))
@@ -306,10 +285,7 @@ initialize_shell_variables (env)
     else
       {
 	if (!tname)
-	  {
-	    char *make_absolute ();
-	    name = make_absolute (shell_name, get_string_value ("PWD"));
-	  }
+	  name = make_absolute (shell_name, get_string_value ("PWD"));
 	else
 	  {
 	    name = full_pathname (tname);
@@ -332,14 +308,15 @@ initialize_shell_variables (env)
   }
 
   /* Make a variable called BASH_VERSION which contains the version info. */
-  {
-    char tt[16];
-    extern char *dist_version;
-    extern int build_version;
+  bind_variable ("BASH_VERSION", shell_version_string ());
 
-    sprintf (tt, "%s.%d-CWRU", dist_version, build_version);
-    bind_variable ("BASH_VERSION", tt);
-  }
+  /* Find out if we're supposed to be in Posix.2 mode via an
+     environment variable. */
+  temp_var = find_variable ("POSIXLY_CORRECT");
+  if (!temp_var)
+    temp_var = find_variable ("POSIX_PEDANTIC");
+  if (temp_var && imported_p (temp_var))
+    sv_strict_posix (temp_var->name);
 
 #if defined (HISTORY)
   /* Set history variables to defaults, and then do whatever we would
@@ -347,15 +324,13 @@ initialize_shell_variables (env)
      that we are remembering commands on the history list. */
   if (remember_on_history)
     {
-      char *tem;
-
       if (posixly_correct)
-	tem = tilde_expand ("~/.sh_history");
+	name = tilde_expand ("~/.sh_history");
       else
-	tem = tilde_expand ("~/.bash_history");
+	name = tilde_expand ("~/.bash_history");
 
-      set_if_not ("HISTFILE", tem);
-      free (tem);
+      set_if_not ("HISTFILE", name);
+      free (name);
 
       set_if_not ("HISTSIZE", "500");
       sv_histsize ("HISTSIZE");
@@ -367,20 +342,19 @@ initialize_shell_variables (env)
 
   /* Handle some "special" variables that we may have inherited from a
      parent shell. */
-  {
-    extern int noclobber;
 
-    noclobber = find_variable ("noclobber") != (SHELL_VAR *)NULL;
-  }
+  noclobber = find_variable ("noclobber") != (SHELL_VAR *)NULL;
+
+  temp_var = find_variable ("IGNOREEOF");
+  if (!temp_var)
+    temp_var = find_variable ("ignoreeof");
+  if (temp_var && imported_p (temp_var))
+    sv_ignoreeof (temp_var->name);
 
 #if defined (HISTORY)
-  {
-    extern int command_oriented_history;
-
-    if (interactive_shell && remember_on_history)
-      command_oriented_history =
-	find_variable ("command_oriented_history") != (SHELL_VAR *)NULL;
-  }
+  if (interactive_shell && remember_on_history)
+    command_oriented_history =
+      find_variable ("command_oriented_history") != (SHELL_VAR *)NULL;
 #endif /* HISTORY */
 
   /* Initialize the dynamic variables, and seed their values. */
@@ -407,10 +381,10 @@ initialize_shell_variables (env)
   non_unsettable ("UID");
 }
 
+void
 adjust_shell_level (change)
      int change;
 {
-  extern int shell_level;
   char *new_level, *old_SHLVL;
   int old_level;
 
@@ -501,7 +475,7 @@ map_over (function, var_hash_table)
   return (list);
 }
 
-int
+static int
 qsort_var_comp (var1, var2)
      SHELL_VAR **var1, **var2;
 {
@@ -513,10 +487,11 @@ qsort_var_comp (var1, var2)
   return (result);
 }
 
+void
 sort_variables (array)
      SHELL_VAR **array;
 {
-  qsort (array, array_len (array), sizeof (SHELL_VAR *), qsort_var_comp);
+  qsort (array, array_len ((char **)array), sizeof (SHELL_VAR *), qsort_var_comp);
 }
 
 /* Create a NULL terminated array of all the shell variables in TABLE. */
@@ -547,6 +522,7 @@ all_shell_functions ()
 }
 
 /* Print VARS to stdout in such a way that they can be read back in. */
+void
 print_var_list (list)
      register SHELL_VAR **list;
 {
@@ -577,6 +553,7 @@ print_vars_no_values (list)
 /* Print the value of a single SHELL_VAR.  No newline is
    output, but the variable is printed in such a way that
    it can be read back in. */
+void
 print_assignment (var)
      SHELL_VAR *var;
 {
@@ -596,6 +573,7 @@ print_assignment (var)
 
 /* Print the value cell of VAR, a shell variable.  Do not print
    the name, nor leading/trailing newline. */
+void
 print_var_value (var)
      SHELL_VAR *var;
 {
@@ -605,6 +583,7 @@ print_var_value (var)
 
 /* Print the function cell of VAR, a shell variable.  Do not
    print the name, nor leading/trailing newline. */
+void
 print_var_function (var)
      SHELL_VAR *var;
 {
@@ -654,10 +633,7 @@ print_var_function (var)
    value of the last assignment. */
 static long seconds_value_assigned = (long)0;
 
-/* Originally defined in shell.c */
-extern time_t shell_start_time;
-
-SHELL_VAR *
+static SHELL_VAR *
 assign_seconds (self, value)
      SHELL_VAR *self;
      char *value;
@@ -667,7 +643,7 @@ assign_seconds (self, value)
   return (self);
 }
 
-SHELL_VAR *
+static SHELL_VAR *
 get_seconds (var)
      SHELL_VAR *var;
 {
@@ -677,8 +653,7 @@ get_seconds (var)
   time_since_start = NOW - shell_start_time;
   p = itos((int) seconds_value_assigned + time_since_start);
 
-  if (var->value)
-    free (var->value);
+  FREE (var->value);
 
   var->attributes |= att_integer;
   var->value = p;
@@ -727,8 +702,8 @@ get_random (var)
 
   rv = brand ();
   p = itos ((int)rv);
-  if (var->value)
-    free (var->value);
+
+  FREE (var->value);
 
   var->attributes |= att_integer;
   var->value = p;
@@ -740,16 +715,29 @@ static SHELL_VAR *
 get_lineno (var)
      SHELL_VAR *var;
 {
-  extern int line_number;
   char *p;
 
   p = itos (line_number);
-  if (var->value)
-    free (var->value);
+  FREE (var->value);
   var->value = p;
   return (var);
 }
 
+#if defined (HISTORY)
+static SHELL_VAR *
+get_histcmd (var)
+     SHELL_VAR *var;
+{
+  char *p;
+
+  p = itos (history_number ());
+  FREE (var->value);
+  var->value = p;
+  return (var);
+}
+#endif
+
+static void
 initialize_dynamic_variables ()
 {
   SHELL_VAR *v;
@@ -765,6 +753,12 @@ initialize_dynamic_variables ()
   v = bind_variable ("LINENO", (char *)NULL);
   v->dynamic_value = get_lineno;
   v->assign_func = (DYNAMIC_FUNC *)NULL;
+
+#if defined (HISTORY)
+  v = bind_variable ("HISTCMD", (char *)NULL);
+  v->dynamic_value = get_histcmd;
+  v->assign_func = (DYNAMIC_FUNC *)NULL;
+#endif
 }
 
 /* How to get a pointer to the shell variable or function named NAME.
@@ -792,8 +786,6 @@ find_variable_internal (name, search_tempenv)
      char *name;
      int search_tempenv;
 {
-  extern int subshell_environment;
-  SHELL_VAR *find_tempenv_variable ();
   SHELL_VAR *var = (SHELL_VAR *)NULL;
 
   /* If explicitly requested, first look in the temporary environment for
@@ -822,9 +814,6 @@ SHELL_VAR *
 find_variable (name)
      char *name;
 {
-  extern int variable_context;
-  extern Function *this_shell_builtin;
-
   return (find_variable_internal
 	  (name, (variable_context || this_shell_builtin || builtin_env)));
 }
@@ -858,7 +847,7 @@ SHELL_VAR *
 make_local_variable (name)
      char *name;
 {
-  SHELL_VAR *new_var, *old_var, *bind_variable ();
+  SHELL_VAR *new_var, *old_var;
   BUCKET_CONTENTS *elt;
 
   /* local foo; local foo;  is a no-op. */
@@ -912,7 +901,7 @@ make_local_variable (name)
         local_variable_stack_size += 8;
       have_local_variables =
         xrealloc (have_local_variables, local_variable_stack_size);
-      bzero (have_local_variables + old_size,
+      bzero ((char *)have_local_variables + old_size,
 	     local_variable_stack_size - old_size);
     }
   have_local_variables[variable_context] = 1;		/* XXX */
@@ -927,6 +916,7 @@ bind_variable (name, value)
      char *name, *value;
 {
   SHELL_VAR *entry = var_lookup (name, shell_variables);
+  BUCKET_CONTENTS *elt;
 
   if (!entry)
     {
@@ -937,7 +927,15 @@ bind_variable (name, value)
       entry->name = savestring (name);
 
       if (value)
-	entry->value = savestring (value);
+	{
+	  if (*value)
+	    entry->value = savestring (value);
+	  else
+	    {
+	      entry->value = xmalloc (1);
+	      entry->value[0] = '\0';
+	    }
+	}
       else
 	entry->value = (char *)NULL;
 
@@ -950,12 +948,8 @@ bind_variable (name, value)
       entry->context = 0;
       entry->prev_context = (SHELL_VAR *)NULL;
 
-      {
-	BUCKET_CONTENTS *elt;
-
-	elt = add_hash_item (savestring (name), shell_variables);
-	elt->data = (char *)entry;
-      }
+      elt = add_hash_item (savestring (name), shell_variables);
+      elt->data = (char *)entry;
     }
   else if (entry->assign_func)
     return ((*(entry->assign_func)) (entry, value));
@@ -977,25 +971,30 @@ bind_variable (name, value)
 	 evaluation done. */
       if (integer_p (entry))
 	{
-	  long val, evalexp ();
+	  long val;
 
 	  val = evalexp (value);
 	  /* We cannot free () entry->value before this; what if the string
 	     we are working is `even=even+2'?  We need the original value
 	     around while we are doing the evaluation to handle any possible
 	     recursion. */
-	  if (entry->value)
-	    free (entry->value);
-
+	  FREE (entry->value);
 	  entry->value = itos (val);
 	}
       else
 	{
-	  if (entry->value)
-	    free (entry->value);
+	  FREE (entry->value);
 
 	  if (value)
-	    entry->value = savestring (value);
+	    {
+	      if (*value)
+		entry->value = savestring (value);
+	      else
+		{
+		  entry->value = xmalloc (1);
+		  entry->value[0] = '\0';
+		}
+	    }
 	  else
 	    entry->value = (char *)NULL;
 	}
@@ -1019,7 +1018,7 @@ dispose_variable (var)
     return;
 
   if (function_p (var))
-    dispose_command (var->value);
+    dispose_command ((COMMAND *)var->value);
   else if (var->value)
     free (var->value);
 
@@ -1153,6 +1152,7 @@ kill_all_local_variables ()
 }
 
 /* Delete the entire contents of the hash table. */
+void
 delete_all_variables (hashed_vars)
      HASH_TABLE *hashed_vars;
 {
@@ -1187,7 +1187,7 @@ delete_all_variables (hashed_vars)
     }
 }
 
-SHELL_VAR *
+static SHELL_VAR *
 new_shell_variable (name)
      char *name;
 {
@@ -1195,7 +1195,7 @@ new_shell_variable (name)
 
   var = (SHELL_VAR *)xmalloc (sizeof (SHELL_VAR));
 
-  bzero (var, sizeof (SHELL_VAR));
+  bzero ((char *)var, sizeof (SHELL_VAR));
   var->name = savestring (name);
   return (var);
 }
@@ -1226,7 +1226,7 @@ bind_function (name, value)
     }
 
   if (entry->value)
-    dispose_command (entry->value);
+    dispose_command ((COMMAND *)entry->value);
 
   if (value)	/* I don't think this can happen anymore */
     entry->value = (char *)copy_command (value);
@@ -1260,7 +1260,7 @@ copy_variable (var)
       copy->name = savestring (var->name);
 
       if (function_p (var))
-	copy->value = (char *)copy_command (var->value);
+	copy->value = (char *)copy_command ((COMMAND *)var->value);
       else if (var->value)
 	copy->value = savestring (var->value);
       else
@@ -1279,6 +1279,7 @@ copy_variable (var)
 
 /* Make the variable associated with NAME be read-only.
    If NAME does not exist yet, create it. */
+void
 set_var_read_only (name)
      char *name;
 {
@@ -1295,6 +1296,7 @@ set_var_read_only (name)
 
 /* Make the function associated with NAME be read-only.
    If NAME does not exist, we just punt, like auto_export code below. */
+void
 set_func_read_only (name)
      char *name;
 {
@@ -1306,6 +1308,7 @@ set_func_read_only (name)
 
 /* Make the variable associated with NAME be auto-exported.
    If NAME does not exist yet, create it. */
+void
 set_var_auto_export (name)
      char *name;
 {
@@ -1322,6 +1325,7 @@ set_var_auto_export (name)
 }
 
 /* Make the function associated with NAME be auto-exported. */
+void
 set_func_auto_export (name)
      char *name;
 {
@@ -1437,10 +1441,10 @@ make_var_array (hashed_vars)
 	  int value_len = strlen (value);
 	  char	*p;
 
-	  p = list[list_index] =  (char *)xmalloc (2 + name_len + value_len);
+	  p = list[list_index] = xmalloc (2 + name_len + value_len);
 	  strcpy (p, var->name);
 	  p[name_len] = '=';
-	  strcpy (&p[name_len + 1], value);
+	  strcpy (p + name_len + 1, value);
 	  list_index++;
 	}
     }
@@ -1460,6 +1464,7 @@ assign_in_env (string)
   int offset = assignment (string);
   char *name = savestring (string);
   char *temp, *value = (char *)NULL;
+  int nlen, vlen;
 
   if (name[offset] == '=')
     {
@@ -1481,9 +1486,14 @@ assign_in_env (string)
   if (!value)
     value = savestring ("");
 
-  temp = (char *)xmalloc (2 + strlen (name) + strlen (value));
-  sprintf (temp, "%s=%s", name, value);
+  nlen = strlen (name);
+  vlen = strlen (value);
+  temp = xmalloc (2 + nlen + vlen);
+  strcpy (temp, name);
+  temp[nlen] = '=';
+  strcpy (temp + nlen + 1, value);
   free (name);
+  free (value);
 
   if (!temporary_env)
     {
@@ -1628,10 +1638,12 @@ qsort_string_compare (s1, s2)
 }
 
 /* Sort ARRAY, a null terminated array of pointers to strings. */
+void
 sort_char_array (array)
      char **array;
 {
-  qsort (array, array_len (array), sizeof (char *), qsort_string_compare);
+  qsort (array, array_len (array), sizeof (char *),
+	 (Function *)qsort_string_compare);
 }
 
 #define ISFUNC(s, o) ((s[o + 1] == '(')  && (s[o + 2] == ')'))
@@ -1673,6 +1685,7 @@ add_or_supercede (assign, array)
    array needs making.  Otherwise, do nothing.  If a shell action could
    change the array that commands receive for their environment, then the
    code should `array_needs_making++'. */
+void
 maybe_make_export_env ()
 {
   register int i;
@@ -1725,17 +1738,18 @@ maybe_make_export_env ()
 }
 
 /* We always put _ in the environment as the name of this command. */
+void
 put_command_name_into_env (command_name)
      char *command_name;
 {
   char *dummy;
 
-  dummy = (char *)xmalloc (4 + strlen (command_name));
+  dummy = xmalloc (4 + strlen (command_name));
 
   /* These three statements replace a call to sprintf */
   dummy[0] = '_';
   dummy[1] = '=';
-  strcpy (&dummy[2], command_name);
+  strcpy (dummy + 2, command_name);
   export_env = add_or_supercede (dummy, export_env);
   free (dummy);
 }
@@ -1747,29 +1761,23 @@ put_command_name_into_env (command_name)
    This screws our scheme.  However, Bash will run on the NeXT using
    the C library getenv (), since right now the only environment variable
    that we care about is HOME, and that is already defined.  */
-#if defined (__STDC__)
-#  define _CONST_HACK const
-#else
-#  define _CONST_HACK
-#endif /* !__STDC__ */
-
 #if !defined (NeXT) && !defined (HPOSF1)
 static char *last_tempenv_value = (char *)NULL;
+extern char **environ;
 
 char *
 getenv (name)
 #if defined (Linux) || defined (__bsdi__)
-     _CONST_HACK char *name;
+     const char *name;
 #else
-     char _CONST_HACK *name;
+     char const *name;
 #endif /* !Linux */
 {
-  SHELL_VAR *var = find_tempenv_variable (name);
+  SHELL_VAR *var = find_tempenv_variable ((char *)name);
 
   if (var)
     {
-      if (last_tempenv_value)
-	free (last_tempenv_value);
+      FREE (last_tempenv_value);
 
       last_tempenv_value = savestring (value_cell (var));
       dispose_variable (var);
@@ -1777,14 +1785,13 @@ getenv (name)
     }
   else if (shell_variables)
     {
-      var = find_variable (name);
+      var = find_variable ((char *)name);
       if (var && exported_p (var))
 	return (value_cell (var));
     }
   else
     {
       register int i, len = strlen (name);
-      extern char **environ;
 
       /* In some cases, s5r3 invokes getenv() before main(); BSD systems
          using gprof also exhibit this behavior.  This means that
@@ -1793,8 +1800,7 @@ getenv (name)
 
       for (i = 0; environ[i]; i++)
 	{
-	  if ((strncmp (environ[i], name, len) == 0) &&
-	      (environ[i][len] == '='))
+	  if ((STREQN (environ[i], name, len)) && (environ[i][len] == '='))
 	    return (environ[i] + len + 1);
 	}
     }
